#ifndef FSGUIPOPUPMENU_IS_INCLUDED
#define FSGUIPOPUPMENU_IS_INCLUDED
/* { */

#include "fsguibase.h"

class FsGuiPopUpMenuItem : private FsGuiObject
{
friend class FsGuiPopUpMenu;

protected:
	int x0,y0,wid,hei;
	YSBOOL checked,isText;
	int id,fsKey;

	int msgWid,msgHei;
	int rightMsgWid,rightMsgHei;

	YsWString wStr;

	mutable FSGUI_DRAWINGTYPE drawingType;
	mutable FsGuiBitmapType bmp;

	YsWString rightAlignedWStr;
	FsGuiBitmapType rightAlignedBmp;



	class FsGuiPopUpMenu *owner;
	class FsGuiPopUpMenu *subMenu;
	void *selectionCallBackParam;
	void (*selectionCallBack)(void *param,class FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *selectedItem);

public:
	FsGuiPopUpMenuItem();
	~FsGuiPopUpMenuItem();

	void Initialize(void);
	class FsGuiPopUpMenu *AddSubMenu(void);
	class FsGuiPopUpMenu *AddSubMenu(int x0,int y0);
	class FsGuiPopUpMenu *GetSubMenu(void);

	const wchar_t *GetString(void) const;

	void SetCheck(YSBOOL checked);
	const YSBOOL GetCheck(void) const;

	const int GetId(void) const;
};

class FsGuiPopUpMenu : private FsGuiControlBase
{
friend class FsGuiPopUpMenuItem;

public:
	using FsGuiControlBase::SetParent;
	using FsGuiControlBase::GetModalDialog;

protected:
	YSBOOL pullDown;

	YsListAllocator <FsGuiPopUpMenuItem> itemAlloc;
	YsListContainer <FsGuiPopUpMenuItem> itemList;
	FsGuiPopUpMenuItem *openSubMenu,*mouseOn;

	FsGuiPopUpMenuItem *owner;

	int x0,y0,wid,hei;

	int lastX,lastY;

	int pMx,pMy;
	YSBOOL pLb,pMb,pRb;
	FsGuiPopUpMenuItem *lastItem;

	mutable YSBOOL needRedraw;

public:
	YsColor bgCol,fgCol;
	YsColor activeBgCol,activeFgCol;
	YsColor frameCol;



	FsGuiPopUpMenu();
	~FsGuiPopUpMenu();

	void SetIsPullDownMenu(YSBOOL isPulLDown);
	YSBOOL IsPullDownMenu(void) const;

	virtual void SetNeedRedraw(YSBOOL needRedraw);
	virtual YSBOOL NeedRedraw(void) const;

	/*! This function really initializes the object. */
	void Initialize(void);

	/*! This function clears menu items, but does not change the location of the pop-up menu. */
	void CleanUp(void);

	FsGuiPopUpMenuItem *GetLastSelectedItem(void);

public:
	FsGuiPopUpMenuItem *AddTextItem(int menuId,int fsKey,const char label[],void (*callback)(void *,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)=NULL,void *contextPtr=NULL);
	FsGuiPopUpMenuItem *AddTextItem(int menuId,int fsKey,const wchar_t wLabel[],void (*callback)(void *,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)=NULL,void *contextPtr=NULL);
	FsGuiPopUpMenuItem *AddBmpItem(int menuId,int fsKey,const YsBitmap &bmp,const wchar_t wLabel[],void (*callback)(void *,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)=NULL,void *contextPtr=NULL);
protected:
	void SetUpMenuItem(
	    FsGuiPopUpMenuItem *item,int id,int fsKey,const wchar_t label[],int msgWid,int msgHei,int rightMsgWid,int rightMsgHei);


public:
	FsGuiPopUpMenuItem *FindItemInRange(int mx,int my) const;
	FsGuiPopUpMenu *GetTopMostMenu(void);

	int GetWidth(void) const;
	int GetHeight(void) const;

	/*! This function sets the width of the menu in pixels. 
	    Size of the pull-down menu may need to be reset to the width of the canvas or the window.
	    This function can be used for that purpose. */
	void SetWidth(int wid);

	/*! This function returns the background color. */
	const YsColor &GetBackgroundColor(void) const;

	/*! This function changes the background color. */
	void SetBackgroundColor(const YsColor &newBackgroundColor);

	YSBOOL IsMouseOnMenu(int mx,int my) const;
	YSRESULT SetMouseState(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	void SetMouseOnItem(int mx,int my);
	void OpenSubMenu(FsGuiPopUpMenuItem *itm);
	FsGuiPopUpMenuItem *GetOpenSubMenu(void);
	void CloseSubMenu(void);
	void CloseAllSubMenu(void);
	YSRESULT MouseButtonPressEvent(FsGuiPopUpMenuItem *&onItem,int mx,int my);
	YSRESULT MouseButtonReleaseEvent(FsGuiPopUpMenuItem *&onItem,int mx,int my);
	YSRESULT KeyIn(int fsKey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);

public:
	void Show(const class FsGuiDialog *excludeFromDrawing=NULL) const;
protected:
	void ShowItem(const FsGuiPopUpMenuItem *item,int rectX0,int rectY0,int rectX1,int rectY1) const;

public:
	virtual void OnOpenSubMenu(FsGuiPopUpMenuItem *item);
	virtual void OnCloseSubMenu(FsGuiPopUpMenuItem *item);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *item);
private:
	void InvokeCallBackIfAssigned(FsGuiPopUpMenuItem *item);
};


/* } */
#endif
