#include <stdio.h>
#include <string.h>

class YsFontBitmap
{
public:
	int wid,hei;
	unsigned char *buf;

	YsFontBitmap();
	~YsFontBitmap();
	void MakeBlank(int wid,int hei);
	void ReadBlackWhiteBmp(const char fn[],int wid,int hei);
};

YsFontBitmap::YsFontBitmap()
{
	wid=0;
	hei=0;
	buf=NULL;
}

YsFontBitmap::~YsFontBitmap()
{
	if(buf!=NULL)
	{
		delete [] buf;
	}
}

void YsFontBitmap::MakeBlank(int wid,int hei)
{
	int i;
	this->wid=wid;
	this->hei=hei;
	int bytePerLineGL;
	bytePerLineGL=((wid+31)/32)*4;
	buf=new unsigned char [bytePerLineGL*hei];
	for(i=0; i<bytePerLineGL*hei; i++)
	{
		buf[i]=0;
	}
}

void YsFontBitmap::ReadBlackWhiteBmp(const char fn[],int wid,int hei)
{
	FILE *fp;  // From 0x003e height*4 bytes.

	fp=fopen(fn,"rb");
	if(fp!=NULL)
	{
		unsigned char bmfHeader[14];
		fread(bmfHeader,1,14,fp);

		unsigned char bmiHeader[40];
		fread(bmiHeader,1,40,fp);

		wid=bmiHeader[4]+(unsigned int)bmiHeader[5]*256;
		hei=bmiHeader[8]+(unsigned int)bmiHeader[9]*256;
		printf("%dx%d\n",wid,hei);

		unsigned char *readBuf;
		int bytePerLineGL,bytePerLineFile;

		bytePerLineFile=((wid+31)/32)*4;
		bytePerLineGL=((wid+31)/32)*4;

		readBuf=new unsigned char [hei*bytePerLineFile];

		fseek(fp,0x3e,SEEK_SET);

		fread(readBuf,1,hei*bytePerLineFile,fp);


		if(buf!=NULL)
		{
			delete [] buf;
		}
		buf=new unsigned char [bytePerLineGL*hei];
		this->wid=wid;
		this->hei=hei;


		int n,y,x;
		n=0;
		for(y=0; y<hei; y++)
		{
			for(x=0; x<bytePerLineGL; x++)
			{
				buf[n++]=~readBuf[y*bytePerLineFile+x];
			}
		}

		delete readBuf;

		fclose(fp);
	}
	else
	{
		printf("Cannot open %s\n",fn);
	}
}


////////////////////////////////////////////////////////////

class YsFontBitmapStore
{
public:
	YsFontBitmap font6x7[256];
	YsFontBitmap font6x8[256];
	YsFontBitmap font6x10[256];
	YsFontBitmap font7x10[256];
	YsFontBitmap font8x8[256];
	YsFontBitmap font8x12[256];
	YsFontBitmap font10x14[256];
	YsFontBitmap font12x16[256];
	YsFontBitmap font16x20[256];
	YsFontBitmap font16x24[256];
	YsFontBitmap font20x28[256];
	YsFontBitmap font20x32[256];
	YsFontBitmap font24x40[256];
	YsFontBitmap font28x44[256];
	YsFontBitmap font32x48[256];

	YsFontBitmap font12P[256];
	YsFontBitmap font14P[256];
	YsFontBitmap font16P[256];

	YsFontBitmapStore();
	void LoadFontAll(void);
	void LoadFont(YsFontBitmap font[],int wid,int hei,const char listFile[]);
	void WriteFontSourceAll(const char fn[]);
	void WriteFontSource(FILE *fp,const YsFontBitmap font[],int wid,int hei);
	void WriteFontSourceP(FILE *fp,const YsFontBitmap font[],int hei);
};



YsFontBitmapStore::YsFontBitmapStore()
{
}

void YsFontBitmapStore::LoadFontAll()
{
	LoadFont(font6x7,8,8,"../6x7Font.txt");
	LoadFont(font6x8,6,8,"../6x8Font.txt");
	LoadFont(font6x10,6,10,"../6x10Font.txt");
	LoadFont(font7x10,7,10,"../7x10Font.txt");
	LoadFont(font8x8,8,8,"../8x8Font.txt");
	LoadFont(font8x12,8,12,"../8x12Font.txt");
	LoadFont(font10x14,10,14,"../10x14Font.txt");
	LoadFont(font12x16,12,16,"../12x16Font.txt");
	LoadFont(font16x20,16,20,"../16x20Font.txt");
	LoadFont(font16x24,16,24,"../16x24Font.txt");
	LoadFont(font20x28,20,28,"../20x28Font.txt");
	LoadFont(font20x32,20,32,"../20x32Font.txt");
	LoadFont(font24x40,20,40,"../24x40Font.txt");
	LoadFont(font28x44,28,44,"../28x44Font.txt");
	LoadFont(font32x48,32,48,"../32x48Font.txt");

	LoadFont(font12P,8,12,"../8x12FontP.txt");
	LoadFont(font14P,10,14,"../10x14FontP.txt");
	LoadFont(font16P,12,16,"../12x16FontP.txt");
}

void YsFontBitmapStore::LoadFont(YsFontBitmap font[],int wid,int hei,const char listFile[])
{
	int i;
	for(i=0; i<32; i++)
	{
		font[i].MakeBlank(wid,hei);
	}

	FILE *fp;
	char str[512],fn[512];
	fp=fopen(listFile,"r");
	for(i=32; i<127; i++)
	{
		if(NULL!=fgets(str,508,fp))
		{
			int l,j;
			l=strlen(str);
			for(j=l-1; j>=0; j--)
			{
				if('\n'!=str[j] && '\r'!=str[j] && ' '!=str[j] && '\t'!=str[j])
				{
					break;
				}
				str[j]=0;
			}

			strcpy(fn,"../");
			strcat(fn,str);
			printf("Read %s\n",fn);
			font[i].ReadBlackWhiteBmp(fn,wid,hei);
		}
		else
		{
			printf("Ran out of the file list (i=%d)\n",i);
		}
	}
	fclose(fp);

	for(i=127; i<256; i++)
	{
		font[i].MakeBlank(wid,hei);
	}
}

void YsFontBitmapStore::WriteFontSourceAll(const char fn[])
{
	FILE *fp=fopen(fn,"w");

	WriteFontSource(fp,font6x7,6,7);
	WriteFontSource(fp,font6x8,6,8);
	WriteFontSource(fp,font6x10,6,10);
	WriteFontSource(fp,font7x10,7,10);
	WriteFontSource(fp,font8x8,8,8);
	WriteFontSource(fp,font8x12,8,12);
	WriteFontSource(fp,font10x14,10,14);
	WriteFontSource(fp,font12x16,12,16);
	WriteFontSource(fp,font16x20,16,20);
	WriteFontSource(fp,font16x24,16,24);
	WriteFontSource(fp,font20x28,20,28);
	WriteFontSource(fp,font20x32,20,32);
	WriteFontSource(fp,font24x40,24,40);
	WriteFontSource(fp,font28x44,28,44);
	WriteFontSource(fp,font32x48,32,48);

	WriteFontSourceP(fp,font12P,12);
	WriteFontSourceP(fp,font14P,14);
	WriteFontSourceP(fp,font16P,16);

	fclose(fp);
}

void YsFontBitmapStore::WriteFontSource(FILE *fp,const YsFontBitmap font[],int wid,int hei)
{
	int ascii;
	for(ascii=0; ascii<256; ascii++)
	{
		int i;
		int bytePerLine,totalByte;;
		bytePerLine=((wid+31)/32)*4;
		totalByte=bytePerLine*hei;

		fprintf(fp,"unsigned char YsFont%dx%d_%03d[%d]={\n",wid,hei,ascii,totalByte);

		// Filter out 1s in the void space >>
		for(i=0; i<hei; i++)
		{
			int lastByte=wid/8;
			int lastByteUsedBit=wid%8;
			if(0!=lastByteUsedBit)
			{
				unsigned char lastByteMask=(255<<(8-lastByteUsedBit));
				font[ascii].buf[i*bytePerLine+lastByte]&=lastByteMask;
			}

			int j;
			for(j=lastByte+1; j<bytePerLine; j++)
			{
				font[ascii].buf[i*bytePerLine+j]=0;
			}
		}
		// Filter out 1s in the void space <<

		for(i=0; i<totalByte; i++)
		{
			fprintf(fp,"0x%02x",font[ascii].buf[i]);
			if(i<totalByte-1)
			{
				fprintf(fp,",");
			}

			if(i%16==15)
			{
				fprintf(fp,"\n");
			}
		}
		fprintf(fp,"};\n");
	}

	fprintf(fp,"unsigned char *YsFont%dx%d[256]={\n",wid,hei);
	for(ascii=0; ascii<256; ascii++)
	{
		fprintf(fp,"YsFont%dx%d_%03d",wid,hei,ascii);
		if(ascii<255)
		{
			fprintf(fp,",");
		}
		if(ascii%4==3)
		{
			fprintf(fp,"\n");
		}
	}
	fprintf(fp,"};\n");
}

void YsFontBitmapStore::WriteFontSourceP(FILE *fp,const YsFontBitmap font[],int hei)
{
	int ascii;
	for(ascii=0; ascii<256; ascii++)
	{
		const int wid=font[ascii].wid;

		int i;
		int bytePerLine,totalByte;;
		bytePerLine=((wid+31)/32)*4;
		totalByte=bytePerLine*hei;

		fprintf(fp,"unsigned char YsFont%dP_%03d[%d]={\n",hei,ascii,totalByte);

		// Filter out 1s in the void space >>
		for(i=0; i<hei; i++)
		{
			int lastByte=wid/8;
			int lastByteUsedBit=wid%8;
			if(0!=lastByteUsedBit)
			{
				unsigned char lastByteMask=(255<<(8-lastByteUsedBit));
				font[ascii].buf[i*bytePerLine+lastByte]&=lastByteMask;
			}

			int j;
			for(j=lastByte+1; j<bytePerLine; j++)
			{
				font[ascii].buf[i*bytePerLine+j]=0;
			}
		}
		// Filter out 1s in the void space <<

		for(i=0; i<totalByte; i++)
		{
			fprintf(fp,"0x%02x",font[ascii].buf[i]);
			if(i<totalByte-1)
			{
				fprintf(fp,",");
			}

			if(i%16==15)
			{
				fprintf(fp,"\n");
			}
		}
		fprintf(fp,"};\n");
	}

	fprintf(fp,"unsigned char *YsFont%dP_Ptn[256]={\n",hei);
	for(ascii=0; ascii<256; ascii++)
	{
		fprintf(fp,"YsFont%dP_%03d",hei,ascii);
		if(ascii<255)
		{
			fprintf(fp,",");
		}
		if(ascii%4==3)
		{
			fprintf(fp,"\n");
		}
	}
	fprintf(fp,"};\n");

	fprintf(fp,"int YsFont%dP_Wid[256]={\n",hei);
	for(ascii=0; ascii<256; ascii++)
	{
		fprintf(fp,"%d",font[ascii].wid);
		if(ascii<255)
		{
			fprintf(fp,",");
		}
		if(ascii%4==3)
		{
			fprintf(fp,"\n");
		}
	}
	fprintf(fp,"};\n");
}

int main(void)
{
	static YsFontBitmapStore fontBmpStore;
printf("%s %d\n",__FUNCTION__,__LINE__);
	fontBmpStore.LoadFontAll();
printf("%s %d\n",__FUNCTION__,__LINE__);
	fontBmpStore.WriteFontSourceAll("../src/ysglfontdata.c");
printf("%s %d\n",__FUNCTION__,__LINE__);
	return 0;
}
