/* ////////////////////////////////////////////////////////////

  File Name: editmenu.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "../fsgui3dapp.h"
#include "../miscdlg.h"
#include <ysgl.h>
#include <ysglbuffer.h>

#include <ysshellextutil.h>
#include <ysshellext_trackingutil.h>
#include <ysshellextedit_topologyutil.h>
#include <ysshellextedit_localop.h>
#include <ysshellext_offsetutil.h>
#include "../file_slideshowdialog.h"
#include "edit_create_airfoil.h"
#include "edit_sweepdialog.h"
#include "edit_mirrorimagedialog.h"
#include "edit_hemmingdialog.h"
#include "edit_rounddialog.h"
#include "edit_createprimitivedialog.h"
#include "edit_projdialog.h"
#include "edit_intersectiondialog.h"
#include "edit_insertdialog.h"
#include "edit_solidofrevolutiondialog.h"
#include "../booldialog.h"

#include "../util/utildlg.h"

#include "../refbmp/refbmpdialog.h"

#include "../dnm/dnmtreedialog.h"

#include "../textresource.h"

#include <ysgebl_gui_extension.h>

void FsGui3DMainCanvas::Edit_ClearUIMode(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	canvas->Edit_ClearUIMode();
}

void FsGui3DMainCanvas::Edit_ClearUIMode(void)
{
	dnmEditMode=YSFALSE;
	Edit_ClearUIModeExceptDNMEditMode();
	guiExtension->OnClearUIMode();
}

void FsGui3DMainCanvas::Edit_ClearUIModeExceptDNMEditMode(void)
{
	if(NULL!=ModeCleanUpCallBack)
	{
		ModeCleanUpCallBack(*this);
	}

	ModeCleanUpCallBack=NULL;
	SpaceKeyCallBack=NULL;
	BackSpaceKeyCallBack=NULL;
	LButtonCallBack=NULL;
	RButtonCallBack=NULL;
	LButtonBoxCallBack=NULL;
	threeDInterface.SetCallBack(NULL,NULL);

	LButtonCallBack=NULL;
	UIBeforeDrawCallBack=NULL;
	UIDrawCallBack2D=NULL;
	UIDrawCallBack3D=NULL;
	UILineVtxBuf.CleanUp();

	threeDInterface.Initialize();

	undoRedoRequireToClearUIMode=YSTRUE;
	deletionRequireToClearUIMode=YSTRUE;

	if(NULL!=numberDlg)
	{
		RemoveDialog(numberDlg);
	}
	if(NULL!=solidOfRevolutionDlg)
	{
		RemoveDialog(solidOfRevolutionDlg);
	}
	if(NULL!=dnmTreeDlg)
	{
		RemoveDialog(dnmTreeDlg);
	}
	if(NULL!=angleSliderDlg)
	{
		RemoveDialog(angleSliderDlg);
	}
	if(NULL!=equilateralDlg)
	{
		RemoveDialog(equilateralDlg);
	}
	if(NULL!=slideShowDlg)
	{
		RemoveDialog(slideShowDlg);
	}
	if(NULL!=parallelogramDlg)
	{
		RemoveDialog(parallelogramDlg);
	}
	if(NULL!=insertVtxDlg)
	{
		RemoveDialog(insertVtxDlg);
	}
	if(NULL!=boolDlg)
	{
		RemoveDialog(boolDlg);
	}
	if(NULL!=chopOffDlg)
	{
		RemoveDialog(chopOffDlg);
	}
	if(NULL!=extendOrShrinkDlg)
	{
		RemoveDialog(extendOrShrinkDlg);
	}
	if(NULL!=createAirfoilDlg)
	{
		RemoveDialog(createAirfoilDlg);
	}
	if(NULL!=refBmpDlg)
	{
		RemoveDialog(refBmpDlg);
	}
	if(NULL!=parallelSweepDlg)
	{
		RemoveDialog(parallelSweepDlg);
	}
	if(NULL!=parallelSweepWithPathDlg)
	{
		RemoveDialog(parallelSweepWithPathDlg);
	}
	if(NULL!=sweepConstEdgeDlg)
	{
		RemoveDialog(sweepConstEdgeDlg);
	}
	if(NULL!=mirrorImageDlg)
	{
		RemoveDialog(mirrorImageDlg);
	}
	if(NULL!=hemmingDlg)
	{
		RemoveDialog(hemmingDlg);
	}
	if(NULL!=roundDlg)
	{
		RemoveDialog(roundDlg);
	}
	if(NULL!=createPrimitiveDlg)
	{
		RemoveDialog(createPrimitiveDlg);
	}
	if(NULL!=projectionDlg)
	{
		RemoveDialog(projectionDlg);
	}

	SetNeedRedraw(YSTRUE);
}

void FsGui3DMainCanvas::Edit_Undo(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	if(NULL!=canvas->slHd && YSOK==canvas->slHd->Undo())
	{
		if(YSTRUE==canvas->undoRedoRequireToClearUIMode)
		{
			Edit_ClearUIMode(appPtr,NULL,NULL);
		}

		canvas->needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_ALL;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

void FsGui3DMainCanvas::Edit_Redo(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	if(NULL!=canvas->slHd && YSOK==canvas->slHd->Redo())
	{
		if(YSTRUE==canvas->undoRedoRequireToClearUIMode)
		{
			Edit_ClearUIMode(appPtr,NULL,NULL);
		}

		canvas->needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_ALL;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Copy(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas.Slhd())
	{
		auto &shl=*canvas.Slhd();

		auto selVtHd=shl.GetSelectedVertex();
		auto selPlHd=shl.GetSelectedPolygon();
		auto selFgHd=shl.GetSelectedFaceGroup();
		auto selCeHd=shl.GetSelectedConstEdge();
		// auto selVlHd=canvas.shl.GetSelectedVolume();

		canvas.cutBuffer.CleanUp();
		canvas.cutBuffer.SetShell((const YsShellExt &)shl);
		canvas.cutBuffer.AddVertex(selVtHd);
		canvas.cutBuffer.AddPolygon(selPlHd);
		canvas.cutBuffer.AddConstEdge(selCeHd);
		canvas.cutBuffer.AddFaceGroup(selFgHd);
		// canvas.cutBuffer.AddVolume(selVlHd);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Paste(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas.Slhd())
	{
		YsShellExtEdit::StopIncUndo incUndo(canvas.Slhd());

		canvas.cutBuffer.Paste(*canvas.Slhd());

		auto selVtHd=canvas.cutBuffer.GetNonDerivedNewVertex();
		auto selPlHd=canvas.cutBuffer.GetNonDerivedNewPolygon();
		auto selCeHd=canvas.cutBuffer.GetNonDerivedNewConstEdge();
		auto selFgHd=canvas.cutBuffer.GetNonDerivedNewFaceGroup();
		auto selVlHd=canvas.cutBuffer.GetNonDerivedNewVolume();

		canvas.Slhd()->SelectVertex(selVtHd);
		canvas.Slhd()->SelectPolygon(selPlHd);
		canvas.Slhd()->SelectConstEdge(selCeHd);
		canvas.Slhd()->SelectFaceGroup(selFgHd);
		// canvas.Slhd()->SelectVolume(selVlHd);

		canvas.needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_ALL;
		canvas.SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_PasteSpecial(void *,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
}

/*static*/void FsGui3DMainCanvas::Edit_DeleteSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	canvas->Edit_DeleteSelection();
}

void FsGui3DMainCanvas::Edit_DeleteSelection(void)
{
	if(NULL!=slHd)
	{
		YsShellExtEdit::StopIncUndo incUndo(slHd);

		YSBOOL worked=YSFALSE;
		YSBOOL remain=YSFALSE;

		YsArray <YsShellExt::FaceGroupHandle> selFgHd;
		slHd->GetSelectedFaceGroup(selFgHd);
		for(auto fgHd : selFgHd)
		{
			if(YSOK!=slHd->DeleteFaceGroup(fgHd))
			{
				remain=YSTRUE;
			}
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_FACEGROUP;
			worked=YSTRUE;
		}

		YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
		slHd->GetSelectedConstEdge(selCeHd);
		for(auto ceHd : selCeHd)
		{
			if(YSOK!=slHd->DeleteConstEdge(ceHd))
			{
				remain=YSTRUE;
			}
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_CONSTEDGE;
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE;
			worked=YSTRUE;
		}

		YsArray <YsShellPolygonHandle> selPlHd;
		slHd->GetSelectedPolygon(selPlHd);
		if(0<selPlHd.GetN())
		{
			slHd->DeleteFaceGroupPolygonMulti(selPlHd);

			if(YSOK!=slHd->DeleteMultiPolygon(selPlHd))
			{
				remain=YSTRUE;
			}
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_POLYGON;
			worked=YSTRUE;
		}


		YsArray <YsShellVertexHandle> selVtHd;
		slHd->GetSelectedVertex(selVtHd);
		if(0<selVtHd.GetN())
		{
			if(YSOK!=slHd->DeleteMultiVertex(selVtHd))
			{
				remain=YSTRUE;
			}
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_VERTEX;
			needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_VERTEX;
			worked=YSTRUE;
		}

		if(YSTRUE==remain)
		{
			MessageDialog(L"Warning",L"Some entities could not be deleted because they are used.");
		}

		if(YSTRUE==worked)
		{
			if(deletionRequireToClearUIMode)
			{
				Edit_ClearUIMode();
			}
			SetNeedRedraw(YSTRUE);
		}
	}
}

/* static */ void FsGui3DMainCanvas::Edit_ForceDeleteSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	auto slHd=canvas.slHd;
	if(NULL!=slHd)
	{
		YsShellExtEdit::StopIncUndo incUndo(slHd);

		// Remove face groups from using volumes

		// Delete face group polygons.
		auto selPlHd=slHd->GetSelectedPolygon();
		if(0<selPlHd)
		{
			if(YSOK==slHd->DeleteFaceGroupPolygonMulti(selPlHd))
			{
				canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_FACEGROUP;
			}
		}

		// Delete polygons and const edges using selected vertices
		auto selVtHd=slHd->GetSelectedVertex();
		for(auto vtHd : selVtHd)
		{
			auto vtPlHd=slHd->FindPolygonFromVertex(vtHd);
			if(0<vtPlHd.GetN())
			{
				slHd->DeleteFaceGroupPolygonMulti(vtPlHd);
				slHd->DeleteMultiPolygon(vtPlHd);
				canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
				canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_POLYGON;
			}

			auto vtCeHd=slHd->FindConstEdgeFromVertex(vtHd);
			if(0<vtCeHd.GetN())
			{
				slHd->DeleteMultiConstEdge(vtCeHd);
				canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_CONSTEDGE;
				canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE;
			}
		}

		Edit_DeleteSelection(appPtr,NULL,NULL);  // Then delete everything selected.
	}
printf("%s %d\n",__FUNCTION__,__LINE__);
}

////////////////////////////////////////////////////////////

PolyCreInsertVertexDialog::PolyCreInsertVertexDialog()
{
	canvas=NULL;
	cancelBtn=NULL;
}

PolyCreInsertVertexDialog::~PolyCreInsertVertexDialog()
{
}

/*static*/ PolyCreInsertVertexDialog *PolyCreInsertVertexDialog::Create(class FsGui3DMainCanvas *canvas)
{
	PolyCreInsertVertexDialog *dlg=new PolyCreInsertVertexDialog;
	dlg->canvas=canvas;
	return dlg;
}

/*static*/ void PolyCreInsertVertexDialog::Delete(PolyCreInsertVertexDialog *ptr)
{
	delete ptr;
}

YSRESULT PolyCreInsertVertexDialog::Make(void)
{
	if(NULL==cancelBtn)
	{
		FsGuiDialog::Initialize();
		insertBtn=AddTextButton(0,FSKEY_ESC,FSGUI_PUSHBUTTON,FSGUI_COMMON_INSERT,YSFALSE);
		selectWhenInsertedBtn=AddTextButton(0,FSKEY_NULL,FSGUI_CHECKBOX,FSGUI_DLG_INSERTVERTEX_ADDVERTEXTOSELECTION,YSFALSE);
		changePointSchemeBtn=AddTextButton(0,FSKEY_NULL,FSGUI_PUSHBUTTON,FSGUI_DLG_INSERTVERTEX_CHANGESCHEME,YSFALSE);
		cancelBtn=AddTextButton(0,FSKEY_ESC,FSGUI_PUSHBUTTON,FSGUI_COMMON_CANCEL,YSFALSE);

		SetArrangeType(FSDIALOG_ARRANGE_TOP_LEFT);
		SetBackgroundAlpha(0.3);
		Fit();
	}

	return YSOK;
}

void PolyCreInsertVertexDialog::RemakeDrawingBuffer(void)
{
	// As required
}

void PolyCreInsertVertexDialog::OnButtonClick(FsGuiButton *btn)
{
	if(btn==changePointSchemeBtn)
	{
		if(FsGui3DInterface::IFTYPE_POINT==canvas->threeDInterface.GetInterfaceType())
		{
			canvas->threeDInterface.ChangeInputPoint1to2(NULL,NULL);
			canvas->SetNeedRedraw(YSTRUE);
		}
		else
		{
			canvas->threeDInterface.ChangeInputPoint2to1();
			canvas->SetNeedRedraw(YSTRUE);
		}
	}
	if(btn==insertBtn)
	{
		canvas->Edit_InsertVertex_SpaceKeyCallBack(*canvas);
	}
	if(btn==cancelBtn)
	{
		canvas->Edit_ClearUIMode();
	}
}

void PolyCreInsertVertexDialog::OnDropListSelChange(FsGuiDropList *,int)
{
}

void PolyCreInsertVertexDialog::OnTextBoxChange(FsGuiTextBox *)
{
}

void PolyCreInsertVertexDialog::OnSliderPositionChange(FsGuiSlider *,const double &,const double &)
{
}

void FsGui3DMainCanvas::Edit_InsertVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		Edit_ClearUIMode(appPtr,NULL,NULL);

		canvas->insertVtxDlg->Make();
		canvas->AddDialog(canvas->insertVtxDlg);
		canvas->ArrangeDialog();

		const YsVec3 focus=canvas->drawEnv.GetViewTarget();
		canvas->threeDInterface.BeginInputPoint1(focus);
		canvas->SetNeedRedraw(YSTRUE);

		canvas->SpaceKeyCallBack=Edit_InsertVertex_SpaceKeyCallBack;
		canvas->UIDrawCallBack2D=NULL;
		canvas->UIDrawCallBack3D=NULL;

		canvas->undoRedoRequireToClearUIMode=YSFALSE;
		canvas->deletionRequireToClearUIMode=YSFALSE;
	}
}

YSRESULT FsGui3DMainCanvas::Edit_InsertVertex_SpaceKeyCallBack(FsGui3DMainCanvas &canvas)
{
	const YsVec3 pos=canvas.threeDInterface.point_pos;
	if(NULL!=canvas.slHd)
	{
		auto vtHd=canvas.slHd->AddVertex(pos);
		if(YSTRUE==canvas.insertVtxDlg->selectWhenInsertedBtn->GetCheck())
		{
			canvas.slHd->AddSelectedVertex(1,&vtHd);
			canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_SELECTED_VERTEX;
		}
		canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_VERTEX;
		canvas.SetNeedRedraw(YSTRUE);
	}
	return YSOK;
}

////////////////////////////////////////////////////////////

void FsGui3DMainCanvas::Edit_InsertPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);

		if(3<=selVtHd.GetN())
		{
			YsShellExtEdit::StopIncUndo incUndo(canvas->slHd);

			YsShellPolygonHandle plHd=canvas->slHd->AddPolygon(selVtHd.GetN(),selVtHd);
			canvas->slHd->SelectVertex(0,NULL);
			canvas->slHd->SetPolygonColor(plHd,canvas->colorPaletteDlg->GetColor());

			canvas->needRemakeDrawingBuffer=NEED_REMAKE_DRAWING_SELECTED_VERTEX|NEED_REMAKE_DRAWING_POLYGON;
			canvas->SetNeedRedraw(YSTRUE);
		}
	}
}

void FsGui3DMainCanvas::Edit_InsertConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);
		canvas->slHd->AddConstEdge(selVtHd,YSFALSE);
		canvas->slHd->SelectVertex(0,NULL);

		canvas->needRemakeDrawingBuffer=NEED_REMAKE_DRAWING_SELECTED_VERTEX|NEED_REMAKE_DRAWING_CONSTEDGE;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

void FsGui3DMainCanvas::Edit_InsertFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas.Slhd())
	{
		auto &shl=*canvas.Slhd();
		auto selPlHd=shl.GetSelectedPolygon();
		if(0<selPlHd.GetN())
		{
			auto fgHd=shl.AddFaceGroup(selPlHd);
			shl.SelectPolygon(0,NULL);

			canvas.needRemakeDrawingBuffer=NEED_REMAKE_DRAWING_SELECTED_POLYGON;
			canvas.SetNeedRedraw(YSTRUE);
		}
	}
}

void FsGui3DMainCanvas::Edit_SmoothShadeSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);

		for(YSSIZE_T idx=0; idx<selVtHd.GetN(); ++idx)
		{
			canvas->slHd->SetVertexRoundFlag(selVtHd[idx],YSTRUE);
		}

		canvas->needRemakeDrawingBuffer=NEED_REMAKE_DRAWING_VERTEX|NEED_REMAKE_DRAWING_POLYGON;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

void FsGui3DMainCanvas::Edit_FlatShadeSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);

		for(YSSIZE_T idx=0; idx<selVtHd.GetN(); ++idx)
		{
			canvas->slHd->SetVertexRoundFlag(selVtHd[idx],YSFALSE);
		}

		canvas->needRemakeDrawingBuffer=NEED_REMAKE_DRAWING_VERTEX|NEED_REMAKE_DRAWING_POLYGON;
		canvas->SetNeedRedraw(YSTRUE);
	}
}



/*static*/ void FsGui3DMainCanvas::Edit_Paint_ByPickingPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		Edit_ClearUIMode(appPtr,NULL,NULL);
		canvas->undoRedoRequireToClearUIMode=YSFALSE;
		canvas->deletionRequireToClearUIMode=YSFALSE;
		canvas->LButtonCallBack=Select_Paint_ByPickingPolygon_LButtonCallBack;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ YSRESULT FsGui3DMainCanvas::Select_Paint_ByPickingPolygon_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL,YSBOOL,YSBOOL,int mx,int my)
{
	if(NULL!=canvas.slHd)
	{
		YsShellPolygonHandle pickedPlHd=canvas.PickedPlHd(mx,my,0,NULL);
		if(NULL!=pickedPlHd)
		{
			YsColor col;
			canvas.slHd->GetColorOfPolygon(col,pickedPlHd);

			const YsColor selCol=canvas.colorPaletteDlg->GetColor();
			if(col!=selCol)
			{
				canvas.slHd->SetPolygonColor(pickedPlHd,selCol);
				canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
				canvas.SetNeedRedraw(YSTRUE);
			}
		}
	}
	return YSOK;
}



/*static*/ void FsGui3DMainCanvas::Edit_Paint_ByPickingFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	canvas->MessageDialog(L"",L"Sorry, not implemented yet.");
}


/*static*/ void FsGui3DMainCanvas::Edit_Paint_SelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellPolygonHandle> selPlHd;
		canvas->slHd->GetSelectedPolygon(selPlHd);

		if(0==selPlHd.GetN())
		{
			canvas->MessageDialog(L"Error",L"No polygon is selected.");
		}
		else
		{
			YSBOOL change=YSFALSE;

			const YsColor selCol=canvas->colorPaletteDlg->GetColor();
			for(YSSIZE_T plIdx=0; plIdx<selPlHd.GetN(); ++plIdx)
			{
				YsColor col;
				canvas->slHd->GetColorOfPolygon(col,selPlHd[plIdx]);

				if(col!=selCol)
				{
					canvas->slHd->SetPolygonColor(selPlHd[plIdx],selCol);
					change=YSTRUE;
				}
			}

			if(YSTRUE==change)
			{
				canvas->needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
				canvas->SetNeedRedraw(YSTRUE);
			}
		}
	}
}


/*static*/ void FsGui3DMainCanvas::Edit_Paint_All(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsShellExtEdit::StopIncUndo incUndo(canvas->slHd);

		const YsColor selCol=canvas->colorPaletteDlg->GetColor();
		for(YsShellPolygonHandle plHd=NULL; YSOK==canvas->slHd->MoveToNextPolygon(plHd); )
		{
			canvas->slHd->SetPolygonColor(plHd,selCol);
		}
		canvas->needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
		canvas->SetNeedRedraw(YSTRUE);
	}
}


/*static*/ void FsGui3DMainCanvas::Edit_Paint_FloodFill_SameColor(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		Edit_ClearUIMode(appPtr,NULL,NULL);
		canvas->undoRedoRequireToClearUIMode=YSFALSE;
		canvas->deletionRequireToClearUIMode=YSFALSE;
		canvas->LButtonCallBack=Edit_Paint_FloodFill_SameColor_LButtonCallBack;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ YSRESULT FsGui3DMainCanvas::Edit_Paint_FloodFill_SameColor_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL,YSBOOL,YSBOOL,int mx,int my)
{
	if(NULL!=canvas.slHd)
	{
		YsShellPolygonHandle pickedPlHd=canvas.PickedPlHd(mx,my,0,NULL);

		if(NULL!=pickedPlHd)
		{
			YsShellExtPolygonFloodFill fill;
			fill.FloodFillSameColor(*((const YsShellExt *)(canvas.slHd)),pickedPlHd);

			const YsColor selCol=canvas.colorPaletteDlg->GetColor();

			YsShellExtEdit::StopIncUndo incUndo(canvas.slHd);

			YSSIZE_T nPl;
			const YsShellPolygonHandle *plHdArray;
			fill.Get(nPl,plHdArray);

			for(YSSIZE_T idx=0; idx<nPl; ++idx)
			{
				canvas.slHd->SetPolygonColor(plHdArray[idx],selCol);
			}

			canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
			canvas.SetNeedRedraw(YSTRUE);

			return YSOK;
		}
	}
	return YSERR;
}


/*static*/ void FsGui3DMainCanvas::Edit_Paint_FloodFill_BoundedByConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		Edit_ClearUIMode(appPtr,NULL,NULL);
		canvas->undoRedoRequireToClearUIMode=YSFALSE;
		canvas->deletionRequireToClearUIMode=YSFALSE;
		canvas->LButtonCallBack=Edit_Paint_FloodFill_BoundedByConstEdge_LButtonCallBack;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ YSRESULT FsGui3DMainCanvas::Edit_Paint_FloodFill_BoundedByConstEdge_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL,YSBOOL,YSBOOL,int mx,int my)
{
	if(NULL!=canvas.slHd)
	{
		YsShellPolygonHandle pickedPlHd=canvas.PickedPlHd(mx,my,0,NULL);

		if(NULL!=pickedPlHd)
		{
			YsShellExtPolygonFloodFill fill;
			fill.FloodFillBoundedByConstEdge(*((const YsShellExt *)(canvas.slHd)),pickedPlHd);

			const YsColor selCol=canvas.colorPaletteDlg->GetColor();

			YsShellExtEdit::StopIncUndo incUndo(canvas.slHd);

			YSSIZE_T nPl;
			const YsShellPolygonHandle *plHdArray;
			fill.Get(nPl,plHdArray);

			for(YSSIZE_T idx=0; idx<nPl; ++idx)
			{
				canvas.slHd->SetPolygonColor(plHdArray[idx],selCol);
			}

			canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_POLYGON;
			canvas.SetNeedRedraw(YSTRUE);

			return YSOK;
		}
	}
	return YSERR;
}



/*static*/ void FsGui3DMainCanvas::Edit_MoveSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		Edit_ClearUIMode(appPtr,NULL,NULL);
		canvas->undoRedoRequireToClearUIMode=YSFALSE;
		canvas->deletionRequireToClearUIMode=YSTRUE;

		canvas->UIDrawCallBack3D=Edit_MoveSelection_DrawCallBack3D;
		canvas->SpaceKeyCallBack=Edit_MoveSelection_SpaceKeyCallBack;

		YsArray <YsShellVertexHandle> allSelVtHd;
		YsShellExtEdit_GetAllVertexInSelection(allSelVtHd,*(canvas->slHd));

		YsVec3 cen;
		canvas->slHd->GetCenterOfVertexGroup(cen,allSelVtHd.GetN(),allSelVtHd);

		canvas->threeDInterface.BeginInputPoint2(cen);
		canvas->threeDInterface.SetCallBack(Edit_MoveSelection_CursorMoved,appPtr);

		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_MoveSelection_DrawCallBack3D(FsGui3DMainCanvas &canvas)
{
	if(NULL!=canvas.slHd)
	{
		struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
		YsGLSLUse3DRenderer(renderer);

		const GLfloat col[4]={0.0f,1.0f,0.0f,1.0f};
		YsGLSLSet3DRendererUniformColorfv(renderer,col);

		YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,canvas.UILineVtxBuf.GetNumVertex(),canvas.UILineVtxBuf);

		YsGLSLEndUse3DRenderer(renderer);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_MoveSelection_CursorMoved(void *appPtr,FsGui3DInterface &)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		canvas->UILineVtxBuf.CleanUp();

		const YsVec3 offset=canvas->threeDInterface.point_pos-canvas->threeDInterface.point_pos_org;

		YsArray <YsShellVertexHandle> allSelVtHd;
		YsShellExtEdit_GetAllVertexInSelection(allSelVtHd,*(canvas->slHd));

		YsHashTable <YsVec3> vtxMap;

		YsGLVertexBuffer &buf=canvas->UILineVtxBuf;
		for(YSSIZE_T idx=0; idx<allSelVtHd.GetN(); ++idx)
		{
			YsVec3 vtPos[2];
			canvas->slHd->GetVertexPosition(vtPos[0],allSelVtHd[idx]);
			vtPos[1]=vtPos[0]+offset;
			buf.AddVertex(vtPos[0]);
			buf.AddVertex(vtPos[1]);

			vtxMap.AddElement(canvas->slHd->GetSearchKey(allSelVtHd[idx]),vtPos[1]);
		}

		for(YSSIZE_T idx=0; idx<allSelVtHd.GetN(); ++idx)
		{
			YsArray <YsShellVertexHandle,16> connVtHd;
			canvas->slHd->GetConnectedVertex(connVtHd,allSelVtHd[idx]);

			YsVec3 vtPos[2];
			canvas->slHd->GetVertexPosition(vtPos[0],allSelVtHd[idx]);
			vtPos[0]+=offset;
			for(YSSIZE_T neiIdx=0; neiIdx<connVtHd.GetN(); ++neiIdx)
			{
				if(YSOK!=vtxMap.FindElement(vtPos[1],canvas->slHd->GetSearchKey(connVtHd[neiIdx])))
				{
					canvas->slHd->GetVertexPosition(vtPos[1],connVtHd[neiIdx]);
				}
				buf.AddVertex(vtPos[0]);
				buf.AddVertex(vtPos[1]);
			}
		}
	}
}

/*static*/ YSRESULT FsGui3DMainCanvas::Edit_MoveSelection_SpaceKeyCallBack(FsGui3DMainCanvas &canvas)
{
	if(NULL!=canvas.slHd)
	{
		const YsVec3 offset=canvas.threeDInterface.point_pos-canvas.threeDInterface.point_pos_org;

		YsArray <YsShellVertexHandle> allSelVtHd;
		YsShellExtEdit_GetAllVertexInSelection(allSelVtHd,*(canvas.slHd));

		YsArray <YsVec3> newPosArray;
		newPosArray.Set(allSelVtHd.GetN(),NULL);
		for(YSSIZE_T idx=0; idx<allSelVtHd.GetN(); ++idx)
		{
			canvas.slHd->GetVertexPosition(newPosArray[idx],allSelVtHd[idx]);
			newPosArray[idx]+=offset;
		}

		canvas.slHd->SetMultiVertexPosition(allSelVtHd,newPosArray);

		canvas.needRemakeDrawingBuffer|=NEED_REMAKE_DRAWING_ALL;
		canvas.SetNeedRedraw(YSTRUE);

		canvas.Edit_ClearUIMode();
	}
	return YSOK;
}

/*static*/ void FsGui3DMainCanvas::Edit_SwapEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);

		YsArray <YsShellPolygonHandle> selPlHd;
		canvas->slHd->GetSelectedPolygon(selPlHd);

		YsShell_SwapInfo swapInfo;
		if(2==selVtHd.GetN())
		{
			if(YSOK!=swapInfo.MakeInfo(*((const YsShell *)canvas->slHd),selVtHd[0],selVtHd[1]))
			{
				canvas->MessageDialog(L"Error",
				    L"Two selected vertices must be used by two triangles.");
				return;
			}
		}
		else if(2==selPlHd.GetN())
		{
			if(YSOK!=swapInfo.MakeInfo(*((const YsShell *)canvas->slHd),selPlHd[0],selPlHd[1]))
			{
				canvas->MessageDialog(L"Error",
				    L"Two selected polygons must be two neighboring triangles.");
				return;
			}
		}
		else
		{
			canvas->MessageDialog(L"Error",
			    L"Two vertices or two neighboring triangles must be selected.");
			return;
		}

		YsShellExtEdit::StopIncUndo incUndo(canvas->slHd);
		canvas->slHd->SetPolygonVertex(swapInfo.triPlHd[0],3,swapInfo.newTriVtHd[0]);
		canvas->slHd->SetPolygonVertex(swapInfo.triPlHd[1],3,swapInfo.newTriVtHd[1]);

		canvas->slHd->SelectVertex(0,NULL);
		canvas->slHd->SelectPolygon(0,NULL);

		canvas->needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_SELECTED_VERTEX|NEED_REMAKE_DRAWING_SELECTED_POLYGON;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_MergePolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellPolygonHandle> selPlHd;
		canvas->slHd->GetSelectedPolygon(selPlHd);

		YsShellPolygonStore selPlHash((const YsShell &)*(canvas->slHd));
		YsArray <unsigned int> selPlKey;
		for(YSSIZE_T idx=0; idx<selPlHd.GetN(); ++idx)
		{
			selPlKey.Append(canvas->slHd->GetSearchKey(selPlHd[idx]));
			selPlHash.AddPolygon(selPlHd[idx]);
		}

		YsShellExtEdit::StopIncUndo incUndo(canvas->slHd);
		for(YSSIZE_T idx=0; idx<selPlKey.GetN(); ++idx)
		{
		RETRY:
			YsShellPolygonHandle seedPlHd=canvas->slHd->FindPolygon(selPlKey[idx]);
			if(NULL!=seedPlHd)
			{
				const int nPlVt=canvas->slHd->GetNumVertexOfPolygon(seedPlHd);
				for(int edIdx=0; edIdx<nPlVt; ++edIdx)
				{
					YsShellPolygonHandle neiPlHd=canvas->slHd->GetNeighborPolygon(seedPlHd,edIdx);
					if(NULL!=neiPlHd && YSTRUE==selPlHash.IsIncluded(neiPlHd))
					{
						YsShell_MergeInfo mergeInfo;
						if(YSOK==mergeInfo.MakeInfo((const YsShell &)*(canvas->slHd),seedPlHd,neiPlHd))
						{
							canvas->slHd->SetPolygonVertex(seedPlHd,mergeInfo.newPlVtHd);
							canvas->slHd->DeleteFaceGroupPolygon(neiPlHd);
							canvas->slHd->DeletePolygon(neiPlHd);
							goto RETRY;
						}
					}
				}
			}
		}

		canvas->slHd->SelectPolygon(0,NULL);

		canvas->needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_SELECTED_VERTEX|NEED_REMAKE_DRAWING_SELECTED_POLYGON;
		canvas->SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_SplitPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle,2> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);

		if(2==selVtHd.GetN())
		{
			YsArray <YsShellPolygonHandle> vtPlHd;
			canvas->slHd->FindPolygonFromVertex(vtPlHd,selVtHd[0]);

			for(YSSIZE_T plIdx=vtPlHd.GetN()-1; 0<=plIdx; --plIdx)
			{
				YsArray <YsShellVertexHandle,4> plVtHd;
				canvas->slHd->GetPolygon(plVtHd,vtPlHd[plIdx]);
				if(YSTRUE!=plVtHd.IsIncluded(selVtHd[1]))
				{
					vtPlHd.DeleteBySwapping(plIdx);
				}
			}

			if(1==vtPlHd.GetN())
			{
				YsShellExtEdit_TopologyUtil topoUtil;
				if(YSOK==topoUtil.SplitPolygon(*(canvas->slHd),vtPlHd[0],selVtHd[0],selVtHd[1]))
				{
					canvas->slHd->SelectVertex(0,NULL);

					canvas->needRemakeDrawingBuffer=(unsigned int)
					    NEED_REMAKE_DRAWING_POLYGON|
					    NEED_REMAKE_DRAWING_SELECTED_VERTEX|NEED_REMAKE_DRAWING_SELECTED_POLYGON;
					canvas->SetNeedRedraw(YSTRUE);
				}
			}
			else
			{
				// Cannot find a unique polygon from two vertices.
			}
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_VertexCollapseMidPoint(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);
		if(2==selVtHd.GetN())
		{
			YsVec3 cen;
			canvas->slHd->GetCenterOfEdge(cen,selVtHd[0],selVtHd[1]);
			canvas->Edit_VertexCollapse(selVtHd[0],selVtHd[1],cen);
		}
		else
		{
			canvas->MessageDialog(L"Error",L"Two vertices must be selected.");
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_VertexCollapseAtoB(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);
		if(2==selVtHd.GetN())
		{
			YsVec3 to;
			canvas->slHd->GetVertexPosition(to,selVtHd[1]);
			canvas->Edit_VertexCollapse(selVtHd[0],selVtHd[1],to);
		}
		else
		{
			canvas->MessageDialog(L"Error",L"Two vertices must be selected.");
		}
	}
}

void FsGui3DMainCanvas::Edit_VertexCollapse(YsShellVertexHandle vtHdToDel,YsShellVertexHandle vtHdToSurvive,const YsVec3 &newPos)
{
	if(NULL!=slHd)
	{
		YsShellExtEdit::StopIncUndo incUndo(slHd);

		if(YSOK==YsShellExtEdit_CollapseEdge(*(YsShellExtEdit *)slHd,vtHdToDel,vtHdToSurvive))
		{
			slHd->SetVertexPosition(vtHdToSurvive,newPos);
			slHd->SelectVertex(0,NULL);

			slHd->SetVertexPosition(vtHdToSurvive,newPos);

			slHd->SelectVertex(0,NULL);

			needRemakeDrawingBuffer=(unsigned int)
			    NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_VERTEX|
			    NEED_REMAKE_DRAWING_SELECTED_VERTEX|NEED_REMAKE_DRAWING_SELECTED_POLYGON|
			    NEED_REMAKE_DRAWING_CONSTEDGE;
			SetNeedRedraw(YSTRUE);
		}
	}
}

////////////////////////////////////////////////////////////

PolyCreHemmingDialog::PolyCreHemmingDialog()
{
}

PolyCreHemmingDialog::~PolyCreHemmingDialog()
{
}

/*static*/ PolyCreHemmingDialog *PolyCreHemmingDialog::Create(class FsGui3DMainCanvas *canvas)
{
	PolyCreHemmingDialog *newDlg=new PolyCreHemmingDialog;
	newDlg->canvas=canvas;
	return newDlg;
}

/*static*/ void PolyCreHemmingDialog::Delete(PolyCreHemmingDialog *ptr)
{
	delete ptr;
}

YSRESULT PolyCreHemmingDialog::Make(void)
{
	Initialize();

	okBtn=AddTextButton(0,FSKEY_SPACE,FSGUI_PUSHBUTTON,FSGUI_COMMON_OK,YSTRUE);
	cancelBtn=AddTextButton(0,FSKEY_ESC,FSGUI_PUSHBUTTON,FSGUI_COMMON_CANCEL,YSFALSE);

	slider=AddHorizontalSlider(0,FSKEY_NULL,50,0.0,1.0,YSTRUE);
	slider->SetPosition(0.4);

	SetArrangeType(FSDIALOG_ARRANGE_TOP_LEFT);
	SetBackgroundAlpha(0.3);
	Fit();

	return YSOK;
}

/*virtual*/ void PolyCreHemmingDialog::OnButtonClick(FsGuiButton *btn)
{
	if(btn==okBtn)
	{
		canvas->Edit_Hemming_AroundSelectedPolygon();
		canvas->Edit_ClearUIMode();
	}
	else if(btn==cancelBtn)
	{
		canvas->Edit_ClearUIMode();
	}
}

/*virtual*/ void PolyCreHemmingDialog::OnSliderPositionChange(FsGuiSlider *,const double &,const double &)
{
}

/*static*/ void FsGui3DMainCanvas::Edit_Hemming_AroundSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	canvas->Edit_ClearUIMode();

	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellPolygonHandle> selPlHd;
		canvas->slHd->GetSelectedPolygon(selPlHd);

		canvas->hemmingDlg->offsetUtil.CleanUp();
		if(YSOK==canvas->hemmingDlg->offsetUtil.SetUpForEdgeConnectedPolygonGroup(*(const YsShell *)canvas->slHd,selPlHd))
		{
			if(YSOK==canvas->hemmingDlg->Make())
			{
				canvas->AddDialog(canvas->hemmingDlg);
				canvas->ArrangeDialog();
				canvas->UIDrawCallBack3D=Edit_Hemming_AroundSelectedPolygon_UIDrawCallBack3D;
			}
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Hemming_AroundOutsideOfSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	canvas->Edit_ClearUIMode();

	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellPolygonHandle> selPlHd;
		canvas->slHd->GetSelectedPolygon(selPlHd);

		canvas->hemmingDlg->offsetUtil.CleanUp();
		if(YSOK==canvas->hemmingDlg->offsetUtil.SetUpForOutsideOfEdgeConnectedPolygonGroup(*(const YsShell *)canvas->slHd,selPlHd))
		{
			if(YSOK==canvas->hemmingDlg->Make())
			{
				canvas->AddDialog(canvas->hemmingDlg);
				canvas->ArrangeDialog();
				canvas->UIDrawCallBack3D=Edit_Hemming_AroundSelectedPolygon_UIDrawCallBack3D;
			}
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Hemming_AroundSelectedPolygon_UIDrawCallBack3D(FsGui3DMainCanvas &canvas)
{
	if(NULL!=canvas.slHd)
	{
		const double t=canvas.hemmingDlg->slider->GetPosition();

		double dist=YsInfinity;
		for(auto newVtx : canvas.hemmingDlg->offsetUtil.newVtxArray)
		{
			const double d=newVtx.maxDist*t;
			dist=YsSmaller(dist,d);
		}

		for(auto &newVtx : canvas.hemmingDlg->offsetUtil.newVtxArray)
		{
			const YsVec3 newVtxPos=newVtx.fromPos+newVtx.dir*dist;
			newVtx.toPos=newVtxPos;
		}

		YsGLVertexBuffer vtxBuf;

		YsArray <YsShellPolygonHandle> selPlHdArray=canvas.hemmingDlg->offsetUtil.srcPlHdArray;

		for(auto plHd : selPlHdArray)
		{
			YsArray <YsVec3,4> plVtPos;
			YsArray <YsShellVertexHandle,4> plVtHd;

			canvas.slHd->GetPolygon(plVtPos,plHd);
			canvas.slHd->GetPolygon(plVtHd,plHd);

			for(YSSIZE_T vtIdx=0; vtIdx<plVtPos.GetN(); ++vtIdx)
			{
				YSSIZE_T newVtxIdx;
				if(YSOK==canvas.hemmingDlg->offsetUtil.vtKeyToNewVtxIndex.FindElement(newVtxIdx,canvas.slHd->GetSearchKey(plVtHd[vtIdx])))
				{
					plVtPos[vtIdx]=canvas.hemmingDlg->offsetUtil.newVtxArray[newVtxIdx].toPos;
				}
			}

			for(YSSIZE_T vtIdx=0; vtIdx<plVtPos.GetN(); ++vtIdx)
			{
				vtxBuf.AddVertex(plVtPos[vtIdx]);
				vtxBuf.AddVertex(plVtPos.GetCyclic(vtIdx+1));
			}

			struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
			YsGLSLUse3DRenderer(renderer);

			const GLfloat edgeCol[4]={0,0,0,1.0f};
			YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
			YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,vtxBuf.GetNumVertex(),vtxBuf);

			YsGLSLEndUse3DRenderer(renderer);
		}
	}
}

void FsGui3DMainCanvas::Edit_Hemming_AroundSelectedPolygon(void)
{
	if(NULL!=slHd)
	{
		const double t=hemmingDlg->slider->GetPosition();

		hemmingDlg->offsetUtil.ApplyByRatio(*(YsShellExtEdit *)slHd,t);

		needRemakeDrawingBuffer=(unsigned int)
		    NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_VERTEX|
		    NEED_REMAKE_DRAWING_SELECTED_POLYGON|NEED_REMAKE_DRAWING_CONSTEDGE;
		SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Hemming_OneSideOfVertexSequence(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	canvas->Edit_ClearUIMode();

	if(NULL!=canvas->slHd)
	{
		auto selVtHd=canvas->slHd->GetSelectedVertex();
		auto selCeHd=canvas->slHd->GetSelectedConstEdge();

		if(0<selVtHd.GetN() && 0<selCeHd.GetN())
		{
			canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_NEEDSELECTVTXORCONSTEDGE);
			return;
		}

		if(0<selCeHd.GetN())
		{
			selVtHd=YsShellExt_TrackingUtil::MakeVertexSequenceFromMultiConstEdge(canvas->slHd->Conv(),selCeHd);
			if(0==selVtHd.GetN())
			{
				canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_CANNOTMAKECONSTEDGECHAIN);
				return;
			}
		}


		YsArray <YsArray <YsShellVertexHandle> > contourVtHd(1,NULL);
		YsArray <YsShellPolygonHandle> selPlHd;

		contourVtHd[0]=selVtHd;
		canvas->slHd->GetSelectedPolygon(selPlHd);

		canvas->hemmingDlg->offsetUtil.CleanUp();
		if(YSOK==canvas->hemmingDlg->offsetUtil.SetUpForOneSideOfVertexSequence(*(const YsShellExt *)canvas->slHd,contourVtHd,selPlHd))
		{
			if(YSOK==canvas->hemmingDlg->Make())
			{
				canvas->AddDialog(canvas->hemmingDlg);
				canvas->ArrangeDialog();
				canvas->UIDrawCallBack3D=Edit_Hemming_AroundSelectedPolygon_UIDrawCallBack3D;
			}
		}
	}
}

////////////////////////////////////////////////////////////

PolyCreProjectionDialog::PolyCreProjectionDialog()
{
	canvas=NULL;
	okBtn=NULL;
	cancelBtn=NULL;
	projDirDrp=NULL;
	projTargetDrp=NULL;
}

PolyCreProjectionDialog::~PolyCreProjectionDialog()
{
}

/*static*/ PolyCreProjectionDialog *PolyCreProjectionDialog::Create(FsGui3DMainCanvas *canvas)
{
	PolyCreProjectionDialog *dlg=new PolyCreProjectionDialog;
	dlg->canvas=canvas;
	return dlg;
}

/*static*/ void PolyCreProjectionDialog::Delete(PolyCreProjectionDialog *ptr)
{
	delete ptr;
}

void PolyCreProjectionDialog::Make(void)
{
	if(NULL==okBtn)
	{
		Initialize();

		okBtn=AddTextButton(0,FSKEY_SPACE,FSGUI_PUSHBUTTON,FSGUI_COMMON_OK,YSTRUE);
		cancelBtn=AddTextButton(0,FSKEY_ESC,FSGUI_PUSHBUTTON,FSGUI_COMMON_CANCEL,YSFALSE);

		AddStaticText(0,FSKEY_NULL,L"Direction",YSTRUE);
		const char * const dirStr[]=
		{
			"Shortest to the projection plane",
			"X-axis",
			"Y-axis",
			"Z-axis",
			"First two vertices",
			"Last two vertices"
		};
		projDirDrp=AddDropList(0,FSKEY_NULL,"Direction",6,dirStr,6,32,32,YSFALSE);

		AddStaticText(0,FSKEY_NULL,L"Target Plane",YSTRUE);
		const char * const targetStr[]=
		{
			"Selected Polygon or Const Edge",
			"First three vertices",
			"Last three vertices",
			"XY-plane",
			"XZ-plane",
			"YZ-plane",
			"Parallel to XY passing through the first vertex",
			"Parallel to XZ passing through the first vertex",
			"Parallel to YZ passing through the first vertex",
			"Parallel to XY passing through the last vertex",
			"Parallel to XZ passing through the last vertex",
			"Parallel to YZ passing through the last vertex"
		};
		projTargetDrp=AddDropList(0,FSKEY_NULL,"Plane",12,targetStr,12,40,40,YSFALSE);
	}

	SetArrangeType(FSDIALOG_ARRANGE_TOP_LEFT);
	SetBackgroundAlpha(0.3);
	Fit();

	RemakeVertexProjectionArray();
}

YSRESULT PolyCreProjectionDialog::RemakeVertexProjectionArray(void)
{
	vertexPositionPair.CleanUp();
	if(NULL!=canvas->slHd)
	{
		YsPlane pln;
		if(YSOK!=CalculateProjectionPlane(pln))
		{
printf("%s %d\n",__FUNCTION__,__LINE__);
			return YSERR;
		}
		YsVec3 dir;
		if(YSOK!=CalculateProjectionDirection(dir,pln))
		{
printf("%s %d\n",__FUNCTION__,__LINE__);
			return YSERR;
		}

		if(fabs(dir*pln.GetNormal())<YsTolerance)
		{
printf("%s %d\n",__FUNCTION__,__LINE__);
printf("%s\n",dir.Txt());
printf("%s\n",pln.GetNormal().Txt());
			return YSERR;
		}

		YsArray <YsShellVertexHandle> selVtHd=GetVertexToBeMoved();

		for(auto vtHd : selVtHd)
		{
			YsVec3 itsc;
			if(YSOK!=pln.GetIntersectionHighPrecision(itsc,canvas->slHd->GetVertexPosition(vtHd),dir))
			{
				vertexPositionPair.CleanUp();
printf("%s %d\n",__FUNCTION__,__LINE__);
				return YSERR;
			}
			vertexPositionPair.Increment();
			vertexPositionPair.Last().a=vtHd;
			vertexPositionPair.Last().b=itsc;
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT PolyCreProjectionDialog::RemakeDrawingBuffer(void)
{
	lineBuf.CleanUp();
	triBuf.CleanUp();

	for(auto pair : vertexPositionPair)
	{
		const YsVec3 vtPos=canvas->slHd->GetVertexPosition(pair.a);
		lineBuf.AddVertex(vtPos);
		lineBuf.AddVertex(pair.b);

		triBuf.AddArrowHeadTriangle(vtPos,pair.b,0.1);
	}

	return YSOK;
}

YSRESULT PolyCreProjectionDialog::CalculateProjectionPlane(YsPlane &pln) const
{
	if(NULL==canvas->slHd)
	{
		return YSERR;
	}

	switch(projTargetDrp->GetSelection())
	{
	default:
		return YSERR;
	case 0: //		"Selected Polygon or Const Edge",
		{
			YsArray <YsShellPolygonHandle> selPlHd;
			canvas->slHd->GetSelectedPolygon(selPlHd);

			YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
			canvas->slHd->GetSelectedConstEdge(selCeHd);

			if(1==selPlHd.GetN() && 0==selCeHd.GetN())
			{
				YsArray <YsVec3> plVtPos;
				canvas->slHd->GetPolygon(plVtPos,selPlHd[0]);
				return pln.MakeBestFitPlane(plVtPos.GetN(),plVtPos);
			}
			else if(0==selPlHd.GetN() && 1==selCeHd.GetN())
			{
				YsArray <YsShellVertexHandle> ceVtHd;
				YSBOOL isLoop;
				canvas->slHd->GetConstEdge(ceVtHd,isLoop,selCeHd[0]);

				YsArray <YsVec3> ceVtPos;
				for(auto vtHd : ceVtHd)
				{
					ceVtPos.Append(canvas->slHd->GetVertexPosition(vtHd));
				}
				return pln.MakeBestFitPlane(ceVtPos.GetN(),ceVtPos);
			}
			else
			{
				return YSERR;
			}
		}
		break;
	case 1: //		"First three vertices",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(3<=selVtHd.GetN())
			{
				const YsVec3 vtPos[3]=
				{
					canvas->slHd->GetVertexPosition(selVtHd[0]),
					canvas->slHd->GetVertexPosition(selVtHd[1]),
					canvas->slHd->GetVertexPosition(selVtHd[2])
				};
				return pln.MakePlaneFromTriangle(vtPos[0],vtPos[1],vtPos[2]);
			}
		}
		break;
	case 2: //		"Last three vertices",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(3<=selVtHd.GetN())
			{
				const YsVec3 vtPos[3]=
				{
					canvas->slHd->GetVertexPosition(selVtHd.GetCyclic(-1)),
					canvas->slHd->GetVertexPosition(selVtHd.GetCyclic(-2)),
					canvas->slHd->GetVertexPosition(selVtHd.GetCyclic(-3))
				};
				return pln.MakePlaneFromTriangle(vtPos[0],vtPos[1],vtPos[2]);
			}
		}
		break;
	case 3: //		"XY-plane",
		pln.Set(YsOrigin(),YsZVec());
		return YSOK;
	case 4: //		"XZ-plane",
		pln.Set(YsOrigin(),YsYVec());
		return YSOK;
	case 5: //		"YZ-plane",
		pln.Set(YsOrigin(),YsXVec());
		return YSOK;
	case 6: //		"Parallel to XY passing through the first vertex",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				const YsVec3 vtPos=canvas->slHd->GetVertexPosition(selVtHd[0]);
				pln.Set(vtPos,YsZVec());
				return YSOK;
			}
		}
		break;
	case 7: //		"Parallel to XZ passing through the first vertex",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				const YsVec3 vtPos=canvas->slHd->GetVertexPosition(selVtHd[0]);
				pln.Set(vtPos,YsYVec());
				return YSOK;
			}
		}
		break;
	case 8: //		"Parallel to YZ passing through the first vertex",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				const YsVec3 vtPos=canvas->slHd->GetVertexPosition(selVtHd[0]);
				pln.Set(vtPos,YsXVec());
				return YSOK;
			}
		}
		break;
	case 9: //		"Parallel to XY passing through the last vertex",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				const YsVec3 vtPos=canvas->slHd->GetVertexPosition(selVtHd.Last());
				pln.Set(vtPos,YsZVec());
				return YSOK;
			}
		}
		break;
	case 10: //		"Parallel to XZ passing through the last vertex",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				const YsVec3 vtPos=canvas->slHd->GetVertexPosition(selVtHd.Last());
				pln.Set(vtPos,YsYVec());
				return YSOK;
			}
		}
		break;
	case 11: //		"Parallel to YZ passing through the last vertex"
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				const YsVec3 vtPos=canvas->slHd->GetVertexPosition(selVtHd.Last());
				pln.Set(vtPos,YsXVec());
				return YSOK;
			}
		}
		break;
	}
	return YSERR;
}

YSRESULT PolyCreProjectionDialog::CalculateProjectionDirection(YsVec3 &dir,const YsPlane &pln) const
{
	switch(projDirDrp->GetSelection())
	{
	case 0: //"Shortest to the projection plane",
		dir=pln.GetNormal();
		return YSOK;
	case 1: //"X-axis",
		dir=YsXVec();
		return YSOK;
	case 2: //"Y-axis",
		dir=YsYVec();
		return YSOK;
	case 3: //"Z-axis",
		dir=YsZVec();
		return YSOK;
	case 4: //"First two vertices"
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(2<=selVtHd.GetN())
			{
				const YsVec3 vtPos[2]=
				{
					canvas->slHd->GetVertexPosition(selVtHd[0]),
					canvas->slHd->GetVertexPosition(selVtHd[1])
				};
				dir=vtPos[1]-vtPos[0];
				return dir.Normalize();
			}
		}
		break;
	case 5: //"Last two vertices",
		{
			YsArray <YsShellVertexHandle> selVtHd;
			canvas->slHd->GetSelectedVertex(selVtHd);
			if(2<=selVtHd.GetN())
			{
				const YsVec3 vtPos[2]=
				{
					canvas->slHd->GetVertexPosition(selVtHd.GetCyclic(-1)),
					canvas->slHd->GetVertexPosition(selVtHd.GetCyclic(-2))
				};
				dir=vtPos[1]-vtPos[0];
				return dir.Normalize();
			}
		}
		break;
	default:
		break;
	}
	return YSERR;
}

YsArray <YsShellVertexHandle> PolyCreProjectionDialog::GetVertexToBeMoved(void) const
{
	auto selVtHd=canvas->slHd->GetSelectedVertex();

	switch(projTargetDrp->GetSelection())
	{
	default:
		break;
	case 1: //		"First three vertices",
		for(int i=0; i<3 && i<selVtHd.GetN(); ++i)
		{
			selVtHd[i]=NULL;
		}
		break;
	case 2: //		"Last three vertices",
		for(int i=selVtHd.GetN()-1; 0<=i && selVtHd.GetN()-3<=i; --i)
		{
			selVtHd[i]=NULL;
		}
		break;
	case 6: //		"Parallel to XY passing through the first vertex",
	case 7: //		"Parallel to XZ passing through the first vertex",
	case 8: //		"Parallel to YZ passing through the first vertex",
		if(0<selVtHd.GetN())
		{
			selVtHd[0]=NULL;
		}
	case 9: //		"Parallel to XY passing through the last vertex",
	case 10: //		"Parallel to XZ passing through the last vertex",
	case 11: //		"Parallel to YZ passing through the last vertex"
		if(0<selVtHd.GetN())
		{
			selVtHd.GetCyclic(-1)=NULL;
		}
		break;
	}

	switch(projDirDrp->GetSelection())
	{
	case 0: //"Shortest to the projection plane",
	case 1: //"X-axis",
	case 2: //"Y-axis",
	case 3: //"Z-axis",
		break;
	case 4: //"First two vertices"
		if(0<selVtHd.GetN())
		{
			selVtHd[0]=NULL;
		}
		break;
	case 5: //"Last two vertices",
		if(0<selVtHd.GetN())
		{
			selVtHd.Last()=NULL;
		}
		break;
	default:
		break;
	}

	YsArray <YsShellVertexHandle> targetVtHd;
	for(auto vtHd : selVtHd)
	{
		if(NULL!=vtHd)
		{
			targetVtHd.Append(vtHd);
		}
	}

	return targetVtHd;
}

void PolyCreProjectionDialog::OnDropListSelChange(FsGuiDropList *,int)
{
	RemakeVertexProjectionArray();
	RemakeDrawingBuffer();
}

void PolyCreProjectionDialog::OnButtonClick(FsGuiButton *btn)
{
	if(okBtn==btn)
	{
		if(YSOK==canvas->Edit_Projection_Project())
		{
			canvas->Edit_ClearUIMode();
		}
	}
	else
	{
		canvas->Edit_ClearUIMode();
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Projection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;

	canvas.Edit_ClearUIMode();

	canvas.undoRedoRequireToClearUIMode=YSTRUE;
	canvas.deletionRequireToClearUIMode=YSTRUE;

	canvas.UIDrawCallBack3D=FsGui3DMainCanvas::Edit_Projection_DrawCallBack3d;;
	canvas.SpaceKeyCallBack=FsGui3DMainCanvas::Edit_Projection_SpaceKeyCallBack;

	canvas.projectionDlg->Make();
	canvas.AddDialog(canvas.projectionDlg);
	canvas.ArrangeDialog();
	canvas.ArrangeDialog();
}

/*static*/ void FsGui3DMainCanvas::Edit_Projection_DrawCallBack3d(FsGui3DMainCanvas &canvas)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsGLSLUse3DRenderer(renderer);

	const GLfloat edgeCol[4]={0,0,0,1.0f};
	YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,canvas.projectionDlg->lineBuf.GetNumVertex(),canvas.projectionDlg->lineBuf);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_TRIANGLES,canvas.projectionDlg->triBuf.GetNumVertex(),canvas.projectionDlg->triBuf);

	YsGLSLEndUse3DRenderer(renderer);
}

/*static*/ YSRESULT FsGui3DMainCanvas::Edit_Projection_SpaceKeyCallBack(FsGui3DMainCanvas &canvas)
{
	if(YSOK==canvas.Edit_Projection_Project())
	{
		canvas.Edit_ClearUIMode();
		return YSOK;
	}
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::Edit_Projection_Project(void)
{
	if(NULL!=slHd && YSOK==projectionDlg->RemakeVertexProjectionArray())
	{
		YsShellExtEdit::StopIncUndo incUndo(slHd);
		for(auto pair : projectionDlg->vertexPositionPair)
		{
			slHd->SetVertexPosition(pair.a,pair.b);
		}
		needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_ALL;
		SetNeedRedraw(YSTRUE);
		return YSOK;
	}
	else
	{
		MessageDialog(L"Error",L"Cannot calculate projection");
	}
	return YSERR;
}

/*static*/ void FsGui3DMainCanvas::Edit_TriangulateSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	YSRESULT res=YSOK;

	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	auto shlPtr=canvas.Slhd();
	if(NULL!=shlPtr)
	{
		auto &shl=*shlPtr;
		auto selPlHd=shl.GetSelectedPolygon();

		YsShellExtEdit::StopIncUndo incUndo(canvas.Slhd());

		for(auto plHd : selPlHd)
		{
			YsArray <YsShellVertexHandle,4> plVtHd;
			shl.GetPolygon(plVtHd,plHd);
			if(3<plVtHd.GetN())
			{
				if(YSOK!=YsShellExtEdit_TriangulatePolygon(shl,plHd))
				{
					res=YSERR;
				}
			}
		}
		shl.SelectPolygon(0,NULL);
	}

	canvas.needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_SELECTED_POLYGON;
	canvas.SetNeedRedraw(YSTRUE);

	if(YSOK!=res)
	{
		canvas.MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_TRIANGULATIONERROR);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_SplitConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas.Slhd())
	{
		auto &shl=*canvas.Slhd();
		auto selVtHd=shl.GetSelectedVertex();
		if(0<selVtHd.GetN())
		{
			YsShellExtEdit::StopIncUndo incUndo(shl);
			for(auto vtHd : selVtHd)
			{
				auto vtCeHd=shl.FindConstEdgeFromVertex(vtHd);
				for(auto ceHd : vtCeHd)
				{
					YsShellExtEdit_SplitConstEdge(shl,ceHd,vtHd);
				}
			}
			shl.SelectConstEdge(0,NULL);

			canvas.needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_CONSTEDGE|NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE;
			canvas.SetNeedRedraw(YSTRUE);
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_InvertConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas.Slhd())
	{
		auto &shl=*canvas.Slhd();

		auto selCeHd=shl.GetSelectedConstEdge();

		YsShellExtEdit::StopIncUndo incUndo(shl);

		for(auto ceHd : selCeHd)
		{
			YSBOOL isLoop;
			YsArray <YsShellVertexHandle> ceVtHd;
			shl.GetConstEdge(ceVtHd,isLoop,ceHd);
			ceVtHd.Invert();
			shl.ModifyConstEdge(ceHd,ceVtHd,isLoop);
		}

		shl.SelectConstEdge(0,NULL);

		canvas.needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_CONSTEDGE|NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE;
		canvas.SetNeedRedraw(YSTRUE);
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_MergeConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(NULL!=canvas.Slhd())
	{
		auto &shl=*canvas.Slhd();

		auto selCeHd=shl.GetSelectedConstEdge();
		if(1<selCeHd.GetN())
		{
			YsShellExt::ConstEdgeHandle modifiedCeHd[1]={YsShellExtEdit_MergeConstEdge(shl,selCeHd,YSTRUE)};
			if(NULL!=modifiedCeHd[0])
			{
				shl.SelectConstEdge(1,modifiedCeHd);
				canvas.needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_CONSTEDGE|NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE;
				canvas.SetNeedRedraw(YSTRUE);
			}
			else
			{
				canvas.MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_CANNOTMAKECONSTEDGECHAIN);
			}
		}
	}
}

