/* ////////////////////////////////////////////////////////////

  File Name: fsgui3dapp.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include <ysport.h>

#include "fsgui3dapp.h"
#include "filename.h"
#include "file_slideshowdialog.h"
#include "config/config.h"
#include "config/configmenu.h"
#include "edit/edit_create_airfoil.h"
#include "refbmp/refbmp.h"
#include "refbmp/refbmpdialog.h"
#include "miscdlg.h"
#include "edit/edit_sweepdialog.h"
#include "edit/edit_mirrorimagedialog.h"
#include "edit/edit_hemmingdialog.h"
#include "edit/edit_rounddialog.h"
#include "edit/edit_createprimitivedialog.h"
#include "edit/edit_projdialog.h"
#include "edit/edit_intersectiondialog.h"
#include "edit/edit_insertdialog.h"
#include "edit/edit_solidofrevolutiondialog.h"
#include "util/utildlg.h"
#include "dnm/dnmtreedialog.h"
#include "help/helpdlg.h"
#include "booldialog.h"

#include "textresource.h"

#include <fsguifiledialog.h>

#include <ysgl.h>

#include "ysshellext_sweeputil.h"


#include <ysgebl_gui_extension.h>


static YSBOOL appMustTerminate=YSFALSE;
static int appSleepInterval=10;
static FsGui3DMainCanvas *appMainCanvas;

////////////////////////////////////////////////////////////

void FsGui3DMainCanvas::DrawingCacheState::Save(const YsShellDnmContainer <YsShellExtEdit>::Node *shl)
{
	slHd=shl;
	whenCached=shl->GetSavePoint();
}

YSBOOL FsGui3DMainCanvas::DrawingCacheState::IsModified(const YsShellDnmContainer <YsShellExtEdit>::Node *shl) const
{
	if(slHd!=shl)
	{
		return YSTRUE;
	}
	if(NULL!=slHd)
	{
		return shl->IsModified(whenCached);
	}
	return YSFALSE;
}

////////////////////////////////////////////////////////////

void FsGui3DMainCanvas::NonManifoldEdgeCache::CacheNonManifoldEdge(const YsShellDnmContainer <YsShellExtEdit>::Node *shl)
{
	lineVtxBuf.CleanUp();
	lineColBuf.CleanUp();

	pointVtxBuf.CleanUp();
	pointPixOffsetBuf.CleanUp();
	pointColBuf.CleanUp();

	const YsColor color[8]=
	{
		YsColor(0.0,0.0,0.0),
		YsColor(0.0,0.0,1.0),
		YsColor(1.0,0.0,0.0),
		YsColor(1.0,0.0,1.0),
		YsColor(0.0,1.0,0.0),
		YsColor(0.0,1.0,1.0),
		YsColor(1.0,1.0,0.0),
		YsColor(1.0,1.0,1.0)
	};


	YsShellEdgeEnumHandle edHd=NULL;
	while(YSOK==shl->MoveToNextEdge(edHd))
	{
		const int nEdPl=shl->GetNumPolygonUsingEdge(edHd);
		if(2!=nEdPl)
		{
			const int colorCode=YsBound(nEdPl,0,7);

			YsShellVertexHandle edVtHd[2];
			shl->GetEdge(edVtHd,edHd);

			YsVec3 edVtPos[2];
			shl->GetVertexPosition(edVtPos[0],edVtHd[0]);
			shl->GetVertexPosition(edVtPos[1],edVtHd[1]);

			lineVtxBuf.AddVertex(edVtPos[0]);
			lineColBuf.AddColor(color[colorCode]);
			lineVtxBuf.AddVertex(edVtPos[1]);
			lineColBuf.AddColor(color[colorCode]);

			for(int i=0; i<2; ++i)
			{
				pointVtxBuf.AddVertex(edVtPos[i]);
				pointColBuf.AddColor(color[colorCode]);
				pointPixOffsetBuf.AddPixOffset(-2.0f,-2.0f);
				pointVtxBuf.AddVertex(edVtPos[i]);
				pointColBuf.AddColor(color[colorCode]);
				pointPixOffsetBuf.AddPixOffset(-2.0f, 6.0f);
				pointVtxBuf.AddVertex(edVtPos[i]);
				pointColBuf.AddColor(color[colorCode]);
				pointPixOffsetBuf.AddPixOffset( 6.0f,-2.0f);
			}
		}
	}

	Save(shl);
}

////////////////////////////////////////////////////////////

FsGui3DMainMenu::FsGui3DMainMenu(class FsGui3DMainCanvas *canvas)
{
	this->canvas=canvas;
}

void FsGui3DMainMenu::OnSelectMenuItem(FsGuiPopUpMenuItem *item)
{
	canvas->OnSelectMenuItem(item);
}

FsGui3DMainCanvas *FsGui3DMainMenu::GetMainCanvas(void) const
{
	return canvas;
}

////////////////////////////////////////////////////////////

FsGui3DMainCanvas::FsGui3DMainCanvas() : threeDInterface(this),viewControlDlg(NULL)
{
	fileAndDirName=PolyCreFileName::Create();
	config=PolyCreConfig::Create();

	slHd=NULL;

	dnmCurrentState=-1;
	dnmEditMode=YSFALSE;

	drawAxis=YSTRUE;
	drawVertex=YSTRUE;
	actuallyDrawVertex=YSTRUE;
	drawConstEdge=YSTRUE;
	actuallyDrawConstEdge=YSTRUE;
	drawPolygonEdge=YSTRUE;
	actuallyDrawPolygonEdge=YSTRUE;
	drawPolygon=YSTRUE;
	actuallyDrawPolygon=YSTRUE;
	drawShrunkPolygon=YSTRUE;
	actuallyDrawShrunkPolygon=YSTRUE;
	drawBackFaceInDifferentColor=YSTRUE;
	drawApplyViewPortToAllShell=YSTRUE;
	drawOtherShell=YSTRUE;
	drawNonManifoldEdge=YSFALSE;
	actuallyDrawNonManifoldEdge=YSFALSE;
	actuallyDrawOtherShell=YSTRUE;
	backFaceColor.SetIntRGB(128,16,16);


	MakeMainMenu();
	SetMenuCheck();

	AddViewControlDialog();
	drawEnv.AttachCanvas(this);
	ModeCleanUpCallBack=NULL;
	SpaceKeyCallBack=NULL;
	BackSpaceKeyCallBack=NULL;
	LButtonCallBack=NULL;
	RButtonCallBack=NULL;
	LButtonBoxCallBack=NULL;
	lBtnDown=YSFALSE;
	lBtnDownMx=0;
	lBtnDownMy=0;
	needRemakeDrawingBuffer=(unsigned int)NEED_REMAKE_DRAWING_ALL;
	needRemakeDrawingBuffer_for_OtherShell=(unsigned int)NEED_REMAKE_DRAWING_ALL;

	undoRedoRequireToClearUIMode=YSTRUE;
	deletionRequireToClearUIMode=YSTRUE;

	vertexPickRadius=5.0f;
	normalVertexMarkerRadius=3.0f;
	roundVertexMarkerRadius=5.0f;
	selectedVertexMarkerRadius=6.0;
	polygonShrinkRatioForPicking=0.7f;

	UIBeforeDrawCallBack=NULL;
	UIDrawCallBack3D=NULL;
	UIDrawCallBack2D=NULL;

	// [Modeless dialogs]
	solidOfRevolutionDlg=PolyCreSolidOfRevolutionDialog::Create(this);
	dnmTreeDlg=PolyCreDnmTreeDialog::Create(this);
	angleSliderDlg=PolyCreAngleSliderDialog::Create(this);
	equilateralDlg=PolyCreEquilateralDialog::Create(this);
	slideShowDlg=PolyCreSlideShowDialog::Create(this);
	parallelogramDlg=PolyCreParallelogramDialog::Create(this);
	insertVtxDlg=PolyCreInsertVertexDialog::Create(this);
	boolDlg=PolyCreBoolDialog::Create(this);
	chopOffDlg=PolyCreChopOffDialog::Create(this);
	colorPaletteDlg=YsColorPaletteDialog::Create();
	refBmpStore=PolyCreReferenceBitmapStore::Create();
	parallelSweepDlg=PolyCreEditParallelSweepDialog::Create();
	parallelSweepWithPathDlg=PolyCreEditParallelSweepWithPathDialog::Create(this);
	sweepConstEdgeDlg=PolyCreEditSweepConstEdgeDialog::Create(this);
	mirrorImageDlg=PolyCreEditCreateMirrorImageDialog::Create();
	sweepInfo=YsShellExt_SweepInfo::Create();
	hemmingDlg=PolyCreHemmingDialog::Create(this);
	roundDlg=PolyCreRoundDialog::Create(this);
	createPrimitiveDlg=PolyCreCreatePrimitiveDialog::Create(this);
	projectionDlg=PolyCreProjectionDialog::Create(this);
	extendOrShrinkDlg=PolyCreExtendOrShrinkDialog::Create(this);

	// [Modal dialogs]
	aboutDlg=PolyCreAboutDialog::Create(this);
	helpDlg=PolyCreHelpDialog::Create(this);
	numberDlg=FsGuiNumberDialog::Create();


	fdlg=new FsGuiFileDialog;
	mbox=new FsGuiMessageBoxDialog;
	createAirfoilDlg=NULL;
	configDlg=NULL;
	refBmpDlg=NULL;

	topBar=new FsGuiStatusBar;
	bottomBar=new FsGuiStatusBar;

	topBar->Make(2,800);
	SetTopStatusBar(topBar);
	bottomBar->Make(2,800);
	SetBottomStatusBar(bottomBar);

	YsColor bgCol=topBar->GetBackgroundColor();
	bgCol.SetAi(160);
	topBar->SetBackgroundColor(bgCol);

	bgCol=bottomBar->GetBackgroundColor();
	bgCol.SetAi(160);
	bottomBar->SetBackgroundColor(bgCol);


	guiExtension=PolyCreGuiExtension::Create();
}

FsGui3DMainCanvas::~FsGui3DMainCanvas()
{
	PolyCreGuiExtension::Delete(guiExtension);


	DeleteMainMenu();

	if(NULL!=numberDlg)
	{
		FsGuiNumberDialog::Delete(numberDlg);
		numberDlg=NULL;
	}
	if(NULL!=solidOfRevolutionDlg)
	{
		PolyCreSolidOfRevolutionDialog::Delete(solidOfRevolutionDlg);
		solidOfRevolutionDlg=NULL;
	}
	if(NULL!=helpDlg)
	{
		PolyCreHelpDialog::Delete(helpDlg);
		helpDlg=NULL;
	}
	if(NULL!=aboutDlg)
	{
		PolyCreAboutDialog::Delete(aboutDlg);
		aboutDlg=NULL;
	}
	if(NULL!=dnmTreeDlg)
	{
		PolyCreDnmTreeDialog::Delete(dnmTreeDlg);
		dnmTreeDlg=NULL;
	}
	if(NULL!=angleSliderDlg)
	{
		PolyCreAngleSliderDialog::Delete(angleSliderDlg);
		angleSliderDlg=NULL;
	}
	if(NULL!=equilateralDlg)
	{
		PolyCreEquilateralDialog::Delete(equilateralDlg);
		equilateralDlg=NULL;
	}
	if(NULL!=slideShowDlg)
	{
		PolyCreSlideShowDialog::Delete(slideShowDlg);
		slideShowDlg=NULL;
	}
	if(NULL!=parallelogramDlg)
	{
		PolyCreParallelogramDialog::Delete(parallelogramDlg);
		parallelogramDlg=NULL;
	}
	if(NULL!=insertVtxDlg)
	{
		PolyCreInsertVertexDialog::Delete(insertVtxDlg);
		insertVtxDlg=NULL;
	}
	if(NULL!=boolDlg)
	{
		PolyCreBoolDialog::Delete(boolDlg);
		boolDlg=NULL;
	}
	if(NULL!=chopOffDlg)
	{
		PolyCreChopOffDialog::Delete(chopOffDlg);
		chopOffDlg=NULL;
	}
	if(NULL!=extendOrShrinkDlg)
	{
		PolyCreExtendOrShrinkDialog::Delete(extendOrShrinkDlg);
		extendOrShrinkDlg=NULL;
	}
	if(NULL!=viewControlDlg)
	{
		FsGui3DViewControlDialog::Delete(viewControlDlg);
		viewControlDlg=NULL;
	}
	if(NULL!=colorPaletteDlg)
	{
		YsColorPaletteDialog::Delete(colorPaletteDlg);
		colorPaletteDlg=NULL;
	}
	if(NULL!=parallelSweepDlg)
	{
		PolyCreEditParallelSweepDialog::Delete(parallelSweepDlg);
		parallelSweepDlg=NULL;
	}
	if(NULL!=parallelSweepWithPathDlg)
	{
		PolyCreEditParallelSweepWithPathDialog::Delete(parallelSweepWithPathDlg);
		parallelSweepWithPathDlg=NULL;
	}
	if(NULL!=sweepConstEdgeDlg)
	{
		PolyCreEditSweepConstEdgeDialog::Delete(sweepConstEdgeDlg);
		sweepConstEdgeDlg=NULL;
	}
	if(NULL!=sweepInfo)
	{
		YsShellExt_SweepInfo::Delete(sweepInfo);
		sweepInfo=NULL;
	}
	if(NULL!=fdlg)
	{
		delete fdlg;
		fdlg=NULL;
	}
	if(NULL!=mbox)
	{
		delete mbox;
		mbox=NULL;
	}

	if(NULL!=createAirfoilDlg)
	{
		YsCreateAirfoilDialog::Delete(createAirfoilDlg);
		createAirfoilDlg=NULL;
	}
	if(NULL!=configDlg)
	{
		PolyCreConfigDialog::Delete(configDlg);
		configDlg=NULL;
	}
	if(NULL!=refBmpStore)
	{
		PolyCreReferenceBitmapStore::Delete(refBmpStore);
		refBmpStore=NULL;
	}
	if(NULL!=refBmpDlg)
	{
		PolyCreRefBmpDialog::Delete(refBmpDlg);
		refBmpDlg=NULL;
	}
	if(NULL!=mirrorImageDlg)
	{
		PolyCreEditCreateMirrorImageDialog::Delete(mirrorImageDlg);
		mirrorImageDlg=NULL;
	}
	if(NULL!=createPrimitiveDlg)
	{
		PolyCreCreatePrimitiveDialog::Delete(createPrimitiveDlg);
		createPrimitiveDlg=NULL;
	}
	if(NULL!=projectionDlg)
	{
		PolyCreProjectionDialog::Delete(projectionDlg);
		projectionDlg=NULL;
	}

	if(NULL!=topBar)
	{
		delete topBar;
		topBar=NULL;
	}
	if(NULL!=bottomBar)
	{
		delete bottomBar;
		bottomBar=NULL;
	}

	PolyCreFileName::Delete(fileAndDirName);
	PolyCreConfig::Delete(config);
}

void FsGui3DMainCanvas::AddViewControlDialog(void)
{
	viewControlDlg=FsGui3DViewControlDialog::Create();
	viewControlDlg->Make(this,&drawEnv);
	this->AddDialog(viewControlDlg);
}

void FsGui3DMainCanvas::DeleteMainMenu(void)
{
	delete mainMenu;
}

void FsGui3DMainCanvas::MessageDialog(const wchar_t title[],const wchar_t msg[])
{
	mbox->Make(title,msg,FSGUI_COMMON_CLOSE,NULL,0,0);
	mbox->SetCloseModalCallBack(NULL);
	AttachModalDialog(mbox);
}

void FsGui3DMainCanvas::MessageDialog(const wchar_t title[],const wchar_t msg[],const wchar_t copyBtn[])
{
	mbox->Make(title,msg,FSGUI_COMMON_CLOSE,NULL,copyBtn,0,0);
	mbox->SetCloseModalCallBack(NULL);
	AttachModalDialog(mbox);
}

void FsGui3DMainCanvas::YesNoDialog(
	    const wchar_t title[],const wchar_t msg[],const wchar_t yes[],const wchar_t no[],
	    void (*closeModalCallBackFunc)(FsGuiControlBase *,FsGuiDialog *,int))
{
	mbox->Make(title,msg,yes,no,1,0);
	mbox->SetCloseModalCallBack(closeModalCallBackFunc);
	AttachModalDialog(mbox);
}

void FsGui3DMainCanvas::OnInterval(void)
{
	{
		int key;
		while(FSKEY_NULL!=(key=FsInkey()))
		{
			if(YSOK!=drawEnv.KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT)))
			{
				if(YSOK!=threeDInterface.KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT)))
				{
					this->KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT));
				}
			}
		}
	}

	{
		int charCode;
		while(0!=(charCode=FsInkeyChar()))
		{
			this->CharIn(charCode);
		}
	}

	{
		int lb,mb,rb,mx,my;
		while(FSMOUSEEVENT_NONE!=FsGetMouseEvent(lb,mb,rb,mx,my))
		{
			if(YSOK!=drawEnv.SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
			{
				if(YSOK!=this->SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
				{
					if(YSOK!=threeDInterface.SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
					{
					}
				}
			}
		}
	}

	int winWid,winHei;
	FsGetWindowSize(winWid,winHei);
	this->SetWindowSize(winWid,winHei,/*autoArrangeDialog=*/YSTRUE);

	drawEnv.SetWindowSize(winWid,winHei);
	drawEnv.SetViewportByTwoCorner(0,0,winWid,winHei);


	if(YSTRUE==NeedRedraw() ||
	   YSTRUE==drawEnv.NeedRedraw() ||
	   YSTRUE==threeDInterface.NeedRedraw() ||
	   0!=FsCheckWindowExposure())
	{
		Draw();
	}
}

void FsGui3DMainCanvas::OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem)
{
	if(NULL!=menuItem)
	{
		YsString sys;
		YsUnicodeToSystemEncoding(sys,menuItem->GetString());
		printf("Selected: %s\n",sys.Txt());
	}
}

YSRESULT FsGui3DMainCanvas::OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt)
{
	switch(fskey)
	{
	case FSKEY_SPACE:
		if(NULL!=SpaceKeyCallBack)
		{
			printf("Reacting to Space key\n");
			return (*SpaceKeyCallBack)(*this);
		}
		break;
	case FSKEY_BS:
		if(NULL!=BackSpaceKeyCallBack)
		{
			printf("Reacting to BackSpace key\n");
			return (*BackSpaceKeyCallBack)(*this);
		}
		break;
	}

	for(YSSIZE_T idx=0; idx<shortCut.GetN(); ++idx)
	{
		if(shortCut[idx].fskey==fskey && shortCut[idx].shift==shift && shortCut[idx].ctrl==ctrl && YSTRUE!=alt)
		{
			printf("Reacting to %s key Shift=%d Ctrl=%d\n",FsKeyCodeToString(fskey),shift,ctrl);
			(*shortCut[idx].callBackFunc)(this,NULL,NULL);
			return YSOK;
		}
	}

	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnLButtonDown(YSBOOL,YSBOOL,YSBOOL,int mx,int my)
{
	if(NULL!=LButtonBoxCallBack)
	{
		lBtnDown=YSTRUE;
		lBtnDownMx=mx;
		lBtnDownMy=my;
	}
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	const int dx=YsAbs(mx-lBtnDownMx);
	const int dy=YsAbs(my-lBtnDownMy);

	lBtnDown=YSFALSE;

	if(NULL!=LButtonBoxCallBack && (16<dx || 16<dy))
	{
		SetNeedRedraw(YSTRUE);
		const int xMin=YsSmaller(mx,lBtnDownMx);
		const int yMin=YsSmaller(my,lBtnDownMy);
		const int xMax=YsGreater(mx,lBtnDownMx);
		const int yMax=YsGreater(my,lBtnDownMy);
		return LButtonBoxCallBack(*this,xMin,yMin,xMax,yMax);
	}
	else if(NULL!=LButtonCallBack)
	{
		return LButtonCallBack(*this,lb,mb,rb,mx,my);
	}
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnRButtonDown(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	if(NULL!=RButtonCallBack)
	{
		return RButtonCallBack(*this,lb,mb,rb,mx,my);
	}
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnMouseMove(YSBOOL,YSBOOL,YSBOOL,int mx,int my,int,int)
{
	if(YSTRUE==lBtnDown)
	{
		moveMx=mx;
		moveMy=my;
		SetNeedRedraw(YSTRUE);
		return YSOK;
	}
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnSaveView(const class FsGui3DViewControlDialog *)
{
	YsString dataDir;
	dataDir.EncodeUTF8 <wchar_t> (GetDataDir());

	YsString def;
	def.MakeFullPathName(dataDir,"*.yvw");

	fdlg->Initialize();
	fdlg->mode=FsGuiFileDialog::MODE_SAVE;
	fdlg->multiSelect=YSFALSE;
	fdlg->title.Set(L"Save Viewpoint");
	fdlg->fileExtensionArray.Append(L".yvw");
	fdlg->defaultFileName.SetUTF8String(def);
	fdlg->SetCloseModalCallBack(OnSaveView_FileSelected);
	AttachModalDialog(fdlg);

	return YSOK;
}

/*static*/ void FsGui3DMainCanvas::OnSaveView_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *fdlg,int returnValue)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(YSOK==(YSRESULT)returnValue && 0<canvas.fdlg->selectedFileArray.GetN())
	{
		YsWString ful(canvas.fdlg->selectedFileArray[0]);

		FILE *fp=YsFileIO::Fopen(canvas.fdlg->selectedFileArray[0],"r");
		if(NULL!=fp)
		{
			fclose(fp);
			canvas.YesNoDialog(FSGUI_DLG_CONFIRMOVERWRITE_TITLE,FSGUI_DLG_CONFIRMOVERWRITE_MESSAGE,FSGUI_COMMON_OK,FSGUI_COMMON_CANCEL,OnSaveView_OverwriteChecked);
		}
		else
		{
			canvas.OnSaveView_OverwriteChecked(appPtr,fdlg,returnValue);
		}
	}
}

/*static*/ void FsGui3DMainCanvas::OnSaveView_OverwriteChecked(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(YSOK==(YSRESULT)returnValue && 0<canvas.fdlg->selectedFileArray.GetN())
	{
		YsWString ful(canvas.fdlg->selectedFileArray[0]);

		FILE *fp=YsFileIO::Fopen(canvas.fdlg->selectedFileArray[0],"w");
		if(NULL!=fp)
		{
			YsTextFileOutputStream outStream(fp);

			int wid,hei;
			FsGetWindowSize(wid,hei);
			canvas.viewControlDlg->GetViewControl().SaveViewpoint(wid,hei,outStream);

			fclose(fp);
		}
	}
}


YSRESULT FsGui3DMainCanvas::OnLoadView(class FsGui3DViewControlDialog *)
{
	YsString dataDir;
	dataDir.EncodeUTF8 <wchar_t> (GetDataDir());

	YsString def;
	def.MakeFullPathName(dataDir,"*.yvw");

	fdlg->Initialize();
	fdlg->mode=FsGuiFileDialog::MODE_OPEN;
	fdlg->multiSelect=YSFALSE;
	fdlg->title.Set(L"Open Viewpoint");
	fdlg->fileExtensionArray.Append(L".yvw");
	fdlg->defaultFileName.SetUTF8String(def);
	fdlg->SetCloseModalCallBack(OnLoadView_FileSelected);
	AttachModalDialog(fdlg);

	return YSOK;
}

/*static*/ void FsGui3DMainCanvas::OnLoadView_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue)
{
	FsGui3DMainCanvas &canvas=*(FsGui3DMainCanvas *)appPtr;
	if(YSOK==(YSRESULT)returnValue && 0<canvas.fdlg->selectedFileArray.GetN())
	{
		YsWString ful(canvas.fdlg->selectedFileArray[0]);

		FILE *fp=YsFileIO::Fopen(canvas.fdlg->selectedFileArray[0],"r");
		if(NULL!=fp)
		{
			YsTextFileInputStream inStream(fp);

			int wid,hei;

			canvas.viewControlDlg->SaveLastView();
			canvas.viewControlDlg->GetViewControl().LoadViewpoint(wid,hei,inStream);

			canvas.SetNeedRedraw(YSTRUE);
			fclose(fp);
		}
	}
}


YsShellVertexHandle FsGui3DMainCanvas::PickedVtHd(int mx,int my,YSSIZE_T nExclude,const YsShellVertexHandle excludeVtHd[]) const
{
	YsShellVertexHandle candidateVtHd=NULL;
	if(NULL!=slHd)
	{
		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		YsKeyStore exclusion;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			exclusion.AddKey(slHd->GetSearchKey(excludeVtHd[idx]));
		}

		double candidateDepth=0.0;

		for(int iter=0; iter<2; ++iter)
		{
			const YsGLVertexBuffer &vtxBuf=(0==iter ? drawBuf.normalVertexPosBuffer : drawBuf.roundVertexPosBuffer);
			const YsArray <int> &idxBuf=(0==iter ? drawBuf.normalVertexIdxBuffer : drawBuf.roundVertexIdxBuffer);

			double z;
			const int pickedIdx=PickedIndexInPointBuffer(z,mx,my,vtxBuf,idxBuf,exclusion);
			if(0<=pickedIdx)
			{
				YsShellVertexHandle vtHd=slHd->FindVertex(pickedIdx);
				if(NULL!=vtHd && (NULL==candidateVtHd || z<candidateDepth))
				{
					candidateVtHd=vtHd;
					candidateDepth=z;
				}
			}
		}
	}
	return candidateVtHd;
}

YsShellVertexHandle FsGui3DMainCanvas::PickedVtHdAmongSelection(int mx,int my) const
{
	if(NULL!=slHd)
	{
		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		YsKeyStore exclusion;

		double z;
		const int pickedIdx=PickedIndexInPointBuffer(z,mx,my,drawBuf.selectedVertexPosBuffer,drawBuf.selectedVertexIdxBuffer,exclusion);
		if(0<=pickedIdx)
		{
			YsShellVertexHandle vtHd=slHd->FindVertex(pickedIdx);
			return vtHd;
		}
	}
	return NULL;
}

YsArray <YsPair <YsShellVertexHandle,double> > FsGui3DMainCanvas::BoxedVtHd(int x0,int y0,int x1,int y1,YSSIZE_T nExclude,const YsShellVertexHandle excludeVtHd[]) const
{
	YsArray <YsPair <YsShellVertexHandle,double> > vtHdDepthPair;
	if(NULL!=slHd)
	{
		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		YsKeyStore exclusion;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			exclusion.AddKey(slHd->GetSearchKey(excludeVtHd[idx]));
		}

		YsArray <YsPair <int,double> > vtKeyDepthPair[2]=
		{
			BoxedIndexInPointBuffer(x0,y0,x1,y1,drawBuf.normalVertexPosBuffer,drawBuf.normalVertexIdxBuffer,exclusion),
			BoxedIndexInPointBuffer(x0,y0,x1,y1,drawBuf.roundVertexPosBuffer,drawBuf.roundVertexIdxBuffer,exclusion)
		};
		for(auto &pairArray : vtKeyDepthPair)
		{
			for(auto pair : pairArray)
			{
				YsShellVertexHandle vtHd=slHd->FindVertex(pair.a);
				if(NULL!=vtHd)
				{
					vtHdDepthPair.Increment();
					vtHdDepthPair.Last().a=vtHd;
					vtHdDepthPair.Last().b=pair.b;
				}
			}
		}
	}
	return vtHdDepthPair;
}

YsArray <YsPair <YsShellVertexHandle,double> > FsGui3DMainCanvas::BoxedVtHdAmongSelection(int x0,int y0,int x1,int y1) const
{
	YsArray <YsPair <YsShellVertexHandle,double> > vtHdDepthPair;
	if(NULL!=slHd)
	{
		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		YsKeyStore exclusion;

		YsArray <YsPair <int,double> > vtKeyDepthPair=
			BoxedIndexInPointBuffer(x0,y0,x1,y1,drawBuf.selectedVertexPosBuffer,drawBuf.selectedVertexIdxBuffer,exclusion);
		for(auto pair : vtKeyDepthPair)
		{
			YsShellVertexHandle vtHd=slHd->FindVertex(pair.a);
			if(NULL!=vtHd)
			{
				vtHdDepthPair.Increment();
				vtHdDepthPair.Last().a=vtHd;
				vtHdDepthPair.Last().b=pair.b;
			}
		}
	}
	return vtHdDepthPair;
}

YsShellPolygonHandle FsGui3DMainCanvas::PickedPlHd(int mx,int my,YSSIZE_T nExclude,const YsShellPolygonHandle excludePlHd[]) const
{
	YsShellPolygonHandle candidatePlHd=NULL;
	if(NULL!=slHd)
	{
		YsKeyStore exclusion;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			exclusion.AddKey(slHd->GetSearchKey(excludePlHd[idx]));
		}

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();
		double candidateDepth=0.0;

		if(YSTRUE==actuallyDrawShrunkPolygon)
		{
			double depth;
			const int idx=PickedIndexInEdgeBuffer(depth,mx,my,drawBuf.shrunkEdgePosBuffer,drawBuf.shrunkEdgeIdxBuffer,exclusion);
			YsShellPolygonHandle plHd;
			if(0<=idx && NULL!=(plHd=slHd->FindPolygon(idx)) && (NULL==candidatePlHd || depth<candidateDepth))
			{
				candidateDepth=depth;
				candidatePlHd=plHd;
			}
		}

		if(YSTRUE==actuallyDrawPolygon)
		{
			for(int i=0; i<5; ++i)
			{
				const YsGLVertexBuffer &vtxBuf=
				   (i==0 ? drawBuf.solidShadedVtxBuffer :
				   (i==1 ? drawBuf.solidUnshadedVtxBuffer :
				   (i==2 ? drawBuf.trspShadedVtxBuffer :
				   (i==3 ? drawBuf.backFaceVtxBuffer :
				          (drawBuf.trspUnshadedVtxBuffer)))));
				const YsArray <int> &idxBuf=
				   (i==0 ? drawBuf.solidShadedIdxBuffer :
				   (i==1 ? drawBuf.solidUnshadedIdxBuffer :
				   (i==2 ? drawBuf.trspShadedIdxBuffer :
				   (i==3 ? drawBuf.backFaceIdxBuffer :
				          (drawBuf.trspUnshadedIdxBuffer)))));

				double depth;
				const int idx=PickedIndexInTriangleBuffer(depth,mx,my,vtxBuf,idxBuf,NULL,&exclusion);
				YsShellPolygonHandle plHd;
				if(0<=idx && NULL!=(plHd=slHd->FindPolygon(idx)) && (NULL==candidatePlHd || depth<candidateDepth))
				{
					candidateDepth=depth;
					candidatePlHd=plHd;
				}
			}
		}
	}
	return candidatePlHd;
}

YsShellPolygonHandle FsGui3DMainCanvas::PickedPlHdAmongSelection(int mx,int my) const
{
	YsShellPolygonHandle candidatePlHd=NULL;
	if(NULL!=slHd)
	{
		YsKeyStore exclusion;

		YsArray <YsShellPolygonHandle> selPlHd;
		slHd->GetSelectedPolygon(selPlHd);
		YsKeyStore selected;
		for(auto plHd : selPlHd)
		{
			selected.AddKey(slHd->GetSearchKey(plHd));
		}

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();
		double candidateDepth=0.0;

		{
			double depth;
			const int idx=PickedIndexInEdgeBuffer(depth,mx,my,drawBuf.selectedPolygonPosBuffer,drawBuf.selectedPolygonIdxBuffer,exclusion);
			YsShellPolygonHandle plHd;
			if(0<=idx && NULL!=(plHd=slHd->FindPolygon(idx)) && (NULL==candidatePlHd || depth<candidateDepth))
			{
				candidateDepth=depth;
				candidatePlHd=plHd;
			}
		}

		if(YSTRUE==actuallyDrawPolygon)
		{
			for(int i=0; i<5; ++i)
			{
				const YsGLVertexBuffer &vtxBuf=
				   (i==0 ? drawBuf.solidShadedVtxBuffer :
				   (i==1 ? drawBuf.solidUnshadedVtxBuffer :
				   (i==2 ? drawBuf.trspShadedVtxBuffer :
				   (i==3 ? drawBuf.backFaceVtxBuffer :
				          (drawBuf.trspUnshadedVtxBuffer)))));
				const YsArray <int> &idxBuf=
				   (i==0 ? drawBuf.solidShadedIdxBuffer :
				   (i==1 ? drawBuf.solidUnshadedIdxBuffer :
				   (i==2 ? drawBuf.trspShadedIdxBuffer :
				   (i==3 ? drawBuf.backFaceIdxBuffer :
				          (drawBuf.trspUnshadedIdxBuffer)))));

				double depth;
				const int idx=PickedIndexInTriangleBuffer(depth,mx,my,vtxBuf,idxBuf,&selected,NULL);
				YsShellPolygonHandle plHd;
				if(0<=idx && NULL!=(plHd=slHd->FindPolygon(idx)) && (NULL==candidatePlHd || depth<candidateDepth))
				{
					candidateDepth=depth;
					candidatePlHd=plHd;
				}
			}
		}
	}
	return candidatePlHd;
}

YsArray <YsPair <YsShellPolygonHandle,double> > FsGui3DMainCanvas::BoxedPlHd(int x0,int y0,int x1,int y1,YSSIZE_T nExclude,const YsShellPolygonHandle excludePlHd[]) const
{
	YsArray <YsPair <YsShellPolygonHandle,double> > plHdDepthPairArray;
	if(NULL!=slHd)
	{
		YsKeyStore exclusion;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			exclusion.AddKey(slHd->GetSearchKey(excludePlHd[idx]));
		}

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		if(YSTRUE==actuallyDrawShrunkPolygon)
		{
			YsArray <YsPair <int,double> > boxed=BoxedIndexInEdgeBuffer(x0,y0,x1,y1,drawBuf.shrunkEdgePosBuffer,drawBuf.shrunkEdgeIdxBuffer,NULL,&exclusion);
			for(auto pair : boxed)
			{
				YsShellPolygonHandle plHd=slHd->FindPolygon(pair.a);
				if(NULL!=plHd)
				{
					plHdDepthPairArray.Increment();
					plHdDepthPairArray.Last().a=plHd;
					plHdDepthPairArray.Last().b=pair.b;
					exclusion.AddKey(pair.a);
				}
			}
		}

		for(auto pair : plHdDepthPairArray)
		{
			exclusion.AddKey(slHd->GetSearchKey(pair.a));
		}

		if(YSTRUE==actuallyDrawPolygon)
		{
			for(int i=0; i<5; ++i)
			{
				const YsGLVertexBuffer &vtxBuf=
				   (i==0 ? drawBuf.solidShadedVtxBuffer :
				   (i==1 ? drawBuf.solidUnshadedVtxBuffer :
				   (i==2 ? drawBuf.trspShadedVtxBuffer :
				   (i==3 ? drawBuf.backFaceVtxBuffer :
				          (drawBuf.trspUnshadedVtxBuffer)))));
				const YsArray <int> &idxBuf=
				   (i==0 ? drawBuf.solidShadedIdxBuffer :
				   (i==1 ? drawBuf.solidUnshadedIdxBuffer :
				   (i==2 ? drawBuf.trspShadedIdxBuffer :
				   (i==3 ? drawBuf.backFaceIdxBuffer :
				          (drawBuf.trspUnshadedIdxBuffer)))));


				YsArray <YsPair <int,double> > boxed=BoxedIndexInTriangleBuffer(x0,y0,x1,y1,vtxBuf,idxBuf,NULL,&exclusion);

				for(auto pair : boxed)
				{
					YsShellPolygonHandle plHd=slHd->FindPolygon(pair.a);
					if(NULL!=plHd)
					{
						plHdDepthPairArray.Increment();
						plHdDepthPairArray.Last().a=plHd;
						plHdDepthPairArray.Last().b=pair.b;
						exclusion.AddKey(pair.a);
					}
				}
			}
		}
	}
	return plHdDepthPairArray;
}

YsArray <YsPair <YsShellPolygonHandle,double> > FsGui3DMainCanvas::BoxedPlHdAmongSelection(int x0,int y0,int x1,int y1) const
{
	YsArray <YsPair <YsShellPolygonHandle,double> > plHdDepthPairArray;
	if(NULL!=slHd)
	{
		YsArray <YsShellPolygonHandle> selPlHd;
		slHd->GetSelectedPolygon(selPlHd);

		YsKeyStore selected,exclusion;
		for(auto plHd : selPlHd)
		{
			selected.AddKey(slHd->GetSearchKey(plHd));
		}

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		{
			YsArray <YsPair <int,double> > boxed=BoxedIndexInEdgeBuffer(x0,y0,x1,y1,drawBuf.selectedPolygonPosBuffer,drawBuf.selectedPolygonIdxBuffer,&selected,&exclusion);
			for(auto pair : boxed)
			{
				YsShellPolygonHandle plHd=slHd->FindPolygon(pair.a);
				if(NULL!=plHd)
				{
					plHdDepthPairArray.Increment();
					plHdDepthPairArray.Last().a=plHd;
					plHdDepthPairArray.Last().b=pair.b;
					exclusion.AddKey(pair.a);
				}
			}
		}

		for(auto pair : plHdDepthPairArray)
		{
			exclusion.AddKey(slHd->GetSearchKey(pair.a));
		}

		if(YSTRUE==actuallyDrawPolygon)
		{
			for(int i=0; i<5; ++i)
			{
				const YsGLVertexBuffer &vtxBuf=
				   (i==0 ? drawBuf.solidShadedVtxBuffer :
				   (i==1 ? drawBuf.solidUnshadedVtxBuffer :
				   (i==2 ? drawBuf.trspShadedVtxBuffer :
				   (i==3 ? drawBuf.backFaceVtxBuffer :
				          (drawBuf.trspUnshadedVtxBuffer)))));
				const YsArray <int> &idxBuf=
				   (i==0 ? drawBuf.solidShadedIdxBuffer :
				   (i==1 ? drawBuf.solidUnshadedIdxBuffer :
				   (i==2 ? drawBuf.trspShadedIdxBuffer :
				   (i==3 ? drawBuf.backFaceIdxBuffer :
				          (drawBuf.trspUnshadedIdxBuffer)))));


				YsArray <YsPair <int,double> > boxed=BoxedIndexInTriangleBuffer(x0,y0,x1,y1,vtxBuf,idxBuf,&selected,&exclusion);

				for(auto pair : boxed)
				{
					YsShellPolygonHandle plHd=slHd->FindPolygon(pair.a);
					if(NULL!=plHd)
					{
						plHdDepthPairArray.Increment();
						plHdDepthPairArray.Last().a=plHd;
						plHdDepthPairArray.Last().b=pair.b;
						exclusion.AddKey(pair.a);
					}
				}
			}
		}
	}
	return plHdDepthPairArray;
}

YsShellExt::ConstEdgeHandle FsGui3DMainCanvas::PickedCeHd(int mx,int my,YSSIZE_T nExclude,const YsShellExt::ConstEdgeHandle excludeCeHd[]) const
{
	YsShellExt::ConstEdgeHandle candidateCeHd=NULL;
	if(NULL!=slHd)
	{
		YsKeyStore exclusion;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			exclusion.AddKey(slHd->GetSearchKey(excludeCeHd[idx]));
		}

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();


		double depth;
		const int idx=PickedIndexInEdgeBuffer(depth,mx,my,drawBuf.constEdgeVtxBuffer,drawBuf.constEdgeIdxBuffer,exclusion);
		if(0<=idx)
		{
			candidateCeHd=slHd->FindConstEdge(idx);
			printf("PickedCeKey=%d\n",slHd->GetSearchKey(candidateCeHd));
		}
	}
	return candidateCeHd;
}

YsShellExt::ConstEdgeHandle FsGui3DMainCanvas::PickedCeHdAmongSelection(int mx,int my) const
{
	if(NULL!=slHd)
	{
		YsKeyStore exclusion;

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		double depth;
		const int idx=PickedIndexInEdgeBuffer(depth,mx,my,drawBuf.selectedConstEdgeVtxBuffer,drawBuf.selectedConstEdgeIdxBuffer,exclusion);
		if(0<=idx)
		{
			return slHd->FindConstEdge(idx);
		}
	}
	return NULL;
}

YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > FsGui3DMainCanvas::BoxedCeHd(int x0,int y0,int x1,int y1,YSSIZE_T nExclude,const YsShellExt::ConstEdgeHandle excludeCeHd[]) const
{
	YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > ceHdDepthPairArray;
	if(NULL!=slHd)
	{
		YsKeyStore exclusion;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			exclusion.AddKey(slHd->GetSearchKey(excludeCeHd[idx]));
		}

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		YsArray <YsPair <int,double> > boxed=BoxedIndexInEdgeBuffer(x0,y0,x1,y1,drawBuf.constEdgeVtxBuffer,drawBuf.constEdgeIdxBuffer,NULL,&exclusion);
		for(auto pair : boxed)
		{
			auto ceHd=slHd->FindConstEdge(pair.a);
			if(NULL!=ceHd)
			{
				ceHdDepthPairArray.Increment();
				ceHdDepthPairArray.Last().a=ceHd;
				ceHdDepthPairArray.Last().b=pair.b;
			}
		}
	}
	return ceHdDepthPairArray;
}

YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > FsGui3DMainCanvas::BoxedCeHdAmongSelection(int x0,int y0,int x1,int y1) const
{
	YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > ceHdDepthPairArray;
	if(NULL!=slHd)
	{
		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

		YsArray <YsPair <int,double> > boxed=BoxedIndexInEdgeBuffer(x0,y0,x1,y1,drawBuf.selectedConstEdgeVtxBuffer,drawBuf.selectedConstEdgeIdxBuffer,NULL,NULL);
		for(auto pair : boxed)
		{
			auto ceHd=slHd->FindConstEdge(pair.a);
			if(NULL!=ceHd)
			{
				ceHdDepthPairArray.Increment();
				ceHdDepthPairArray.Last().a=ceHd;
				ceHdDepthPairArray.Last().b=pair.b;
			}
		}
	}
	return ceHdDepthPairArray;
}

YsShellExt::FaceGroupHandle FsGui3DMainCanvas::PickedFgHd(int mx,int my,int nExclude,const YsShellExt::FaceGroupHandle excludeFgHd[]) const
{
	YsShellExt::FaceGroupHandle candidateFgHd=NULL;
	if(NULL!=slHd)
	{
		YsKeyStore excludeFgKey;
		for(YSSIZE_T idx=0; idx<nExclude; ++idx)
		{
			excludeFgKey.AddKey(slHd->GetSearchKey(excludeFgHd[idx]));
		}



		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();
		double candidateDepth=0.0;

		if(YSTRUE==actuallyDrawShrunkPolygon)
		{
			for(YSSIZE_T edIdx=0; edIdx<=drawBuf.shrunkEdgePosBuffer.GetNumVertex()-2; edIdx+=2)
			{
				YsVec3 edVtPos[2]=
				{
					YsVec3(drawBuf.shrunkEdgePosBuffer[edIdx*3  ],drawBuf.shrunkEdgePosBuffer[edIdx*3+1],drawBuf.shrunkEdgePosBuffer[edIdx*3+2]),
					YsVec3(drawBuf.shrunkEdgePosBuffer[edIdx*3+3],drawBuf.shrunkEdgePosBuffer[edIdx*3+4],drawBuf.shrunkEdgePosBuffer[edIdx*3+5]),
				};

				double depth;
				if(YSTRUE==drawEnv.IsLinePicked(depth,mx,my,10,edVtPos[0],edVtPos[1]))
				{
					YsShellPolygonHandle plHd=slHd->FindPolygon(drawBuf.shrunkEdgeIdxBuffer[edIdx/2]);
					YsShellExt::FaceGroupHandle fgHd=slHd->FindFaceGroupFromPolygon(plHd);

					if(NULL==candidateFgHd || depth<candidateDepth)
					{
						if(NULL!=fgHd && YSTRUE!=excludeFgKey.IsIncluded(slHd->GetSearchKey(fgHd)))
						{
							candidateFgHd=fgHd;
							candidateDepth=depth;
						}
					}
				}
			}
		}

		if(YSTRUE==actuallyDrawPolygon)
		{
			for(int i=0; i<5; ++i)
			{
				const YsGLVertexBuffer &vtxBuf=
				   (i==0 ? drawBuf.solidShadedVtxBuffer :
				   (i==1 ? drawBuf.solidUnshadedVtxBuffer :
				   (i==2 ? drawBuf.trspShadedVtxBuffer :
				   (i==3 ? drawBuf.backFaceVtxBuffer :
				          (drawBuf.trspUnshadedVtxBuffer)))));
				const YsArray <int> &idxBuf=
				   (i==0 ? drawBuf.solidShadedIdxBuffer :
				   (i==1 ? drawBuf.solidUnshadedIdxBuffer :
				   (i==2 ? drawBuf.trspShadedIdxBuffer :
				   (i==3 ? drawBuf.backFaceIdxBuffer :
				          (drawBuf.trspUnshadedIdxBuffer)))));

				for(YSSIZE_T plIdx=0; plIdx<=vtxBuf.GetNumVertex()-3; plIdx+=3)
				{
					const YsVec3 triVtPos[3]=
					{
						YsVec3(vtxBuf[plIdx*3  ],vtxBuf[plIdx*3+1],vtxBuf[plIdx*3+2]),
						YsVec3(vtxBuf[plIdx*3+3],vtxBuf[plIdx*3+4],vtxBuf[plIdx*3+5]),
						YsVec3(vtxBuf[plIdx*3+6],vtxBuf[plIdx*3+7],vtxBuf[plIdx*3+8])
					};

					double depth;
					if(YSTRUE==drawEnv.IsPolygonPicked(depth,mx,my,3,triVtPos))
					{
						if(NULL==candidateFgHd || depth<candidateDepth)
						{
							YsShellPolygonHandle plHd=slHd->FindPolygon(idxBuf[plIdx/3]);
							YsShellExt::FaceGroupHandle fgHd=slHd->FindFaceGroupFromPolygon(plHd);
							if(NULL!=fgHd && YSTRUE!=excludeFgKey.IsIncluded(slHd->GetSearchKey(fgHd)))
							{
								candidateFgHd=fgHd;
								candidateDepth=depth;
							}
						}
					}
				}
			}
		}
	}
	return candidateFgHd;
}

YsShellExt::FaceGroupHandle FsGui3DMainCanvas::PickedFgHdAmongSelection(int mx,int my) const
{
	YsShellExt::FaceGroupHandle candidateFgHd=NULL;
	if(NULL!=slHd)
	{
		YsKeyStore excludeFgKey;

		YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();
		double candidateDepth=0.0;

		auto index=PickedIndexInTriangleBuffer(candidateDepth,mx,my,drawBuf.selectedFaceGroupVtxBuffer,drawBuf.selectedFaceGroupIdxBuffer,NULL,NULL);;
		candidateFgHd=slHd->FindFaceGroup(index);
	}
	return candidateFgHd;
}

const YsShellDnmContainer <YsShellExtEdit>::Node *FsGui3DMainCanvas::PickedSlHd(int mx,int my,YSSIZE_T nExclude,const YsShellDnmContainer <YsShellExtEdit>::Node *excludeSlHd[]) const
{
	YsArray <const YsShellDnmContainer <YsShellExtEdit>::Node *> allNode;
	shlGrp.GetNodePointerAll(allNode);

	if(0<allNode.GetN())
	{
		double candidateDepth=0.0;
		const YsShellDnmContainer <YsShellExtEdit>::Node *candidateSlHd=NULL;

		for(YSSIZE_T slIdx=0; slIdx<allNode.GetN(); ++slIdx)
		{
			if(YSTRUE==YsIsIncluded(nExclude,excludeSlHd,allNode[slIdx]))
			{
				continue;
			}

			YsShellExtDrawingBuffer &drawBuf=allNode[slIdx]->GetDrawingBuffer();
			YsGLVertexBuffer &vtxBuf=drawBuf.normalEdgePosBuffer;

			for(YSSIZE_T edIdx=0; edIdx<=vtxBuf.GetNumVertex()-2; edIdx+=2)
			{
				YsVec3 edVtPos[2]=
				{
					YsVec3(vtxBuf[edIdx*3  ],vtxBuf[edIdx*3+1],vtxBuf[edIdx*3+2]),
					YsVec3(vtxBuf[edIdx*3+3],vtxBuf[edIdx*3+4],vtxBuf[edIdx*3+5]),
				};

				double depth;
				if(YSTRUE==drawEnv.IsLinePicked(depth,mx,my,10,edVtPos[0],edVtPos[1]))
				{
					if(NULL==candidateSlHd || depth<candidateDepth)
					{
						candidateSlHd=allNode[slIdx];
						candidateDepth=depth;
					}
				}
			}
		}

		return candidateSlHd;
	}

	return NULL;
}

YsShellDnmContainer <YsShellExtEdit>::Node *FsGui3DMainCanvas::PickedSlHd(int mx,int my,YSSIZE_T nExclude,const YsShellDnmContainer <YsShellExtEdit>::Node *excludeSlHd[])
{
	const FsGui3DMainCanvas *constPtr=this;
	const YsShellDnmContainer <YsShellExtEdit>::Node *pickedSlHd=constPtr->PickedSlHd(mx,my,nExclude,excludeSlHd);
	return (YsShellDnmContainer <YsShellExtEdit>::Node *)pickedSlHd;
}

int FsGui3DMainCanvas::PickedIndexInPointBuffer(double &z,int mx,int my,const YsGLVertexBuffer &pointVtxBuffer,const YsArray <int> &pointIdxBuffer,const YsKeyStore &exclusion) const
{
	int candidateIdx=-1;
	double candidateDepth=0.0;

	for(YSSIZE_T vtIdx=0; vtIdx<pointVtxBuffer.GetNumVertex(); vtIdx+=3)
	{
		if(YSTRUE!=exclusion.IsIncluded(pointIdxBuffer[vtIdx/3]))
		{
			YsVec3 vtPos(pointVtxBuffer[vtIdx*3],pointVtxBuffer[vtIdx*3+1],pointVtxBuffer[vtIdx*3+2]);

			double z;
			if(YSTRUE==drawEnv.IsPointPicked(z,mx,my,(int)vertexPickRadius,vtPos))
			{
				// Why vtIdx*3 and vtIdx/3?  See ysshellextdrawbuffer.h  It is due to unreliability of point-sprite.
				if(0>candidateIdx || candidateDepth>z)
				{
					candidateIdx=pointIdxBuffer[vtIdx/3];
					candidateDepth=z;
				}
			}
		}
	}

	z=candidateDepth;
	return candidateIdx;
}

int FsGui3DMainCanvas::PickedIndexInEdgeBuffer(double &z,int mx,int my,const YsGLVertexBuffer &edgeVtxBuffer,const YsArray <int> &edgeIdxBuffer,const YsKeyStore &exclusion) const
{
	int candidateIndex=-1;
	double candidateDepth=0.0;

	for(YSSIZE_T edIdx=0; edIdx<=edgeVtxBuffer.GetNumVertex()-2; edIdx+=2)
	{
		if(YSTRUE!=exclusion.IsIncluded((YSHASHKEY)edgeIdxBuffer[edIdx/2]))
		{
			YsVec3 edVtPos[2]=
			{
				YsVec3(edgeVtxBuffer[edIdx*3  ],edgeVtxBuffer[edIdx*3+1],edgeVtxBuffer[edIdx*3+2]),
				YsVec3(edgeVtxBuffer[edIdx*3+3],edgeVtxBuffer[edIdx*3+4],edgeVtxBuffer[edIdx*3+5]),
			};

			double depth;
			if(YSTRUE==drawEnv.IsLinePicked(depth,mx,my,10,edVtPos[0],edVtPos[1]))
			{
				if(candidateIndex<0 || depth<candidateDepth)
				{
					candidateIndex=edgeIdxBuffer[edIdx/2];
					candidateDepth=depth;
				}
			}
		}
	}

	z=candidateDepth;
	return candidateIndex;
}

int FsGui3DMainCanvas::PickedIndexInTriangleBuffer(
    double &z,int mx,int my,const YsGLVertexBuffer &triVtxBuffer,const YsArray <int> &triIdxBuffer,const YsKeyStore *limit,const YsKeyStore *exclusion) const
{
	int candidateIndex=-1;
	double candidateDepth=YsInfinity;

	for(YSSIZE_T plIdx=0; plIdx<=triVtxBuffer.GetNumVertex()-3; plIdx+=3)
	{
		if(NULL!=limit && YSTRUE!=limit->IsIncluded(triIdxBuffer[plIdx/3]))
		{
			continue;
		}

		if(NULL==exclusion || YSTRUE!=exclusion->IsIncluded(triIdxBuffer[plIdx/3]))
		{
			const YsVec3 triVtPos[3]=
			{
				YsVec3(triVtxBuffer[plIdx*3  ],triVtxBuffer[plIdx*3+1],triVtxBuffer[plIdx*3+2]),
				YsVec3(triVtxBuffer[plIdx*3+3],triVtxBuffer[plIdx*3+4],triVtxBuffer[plIdx*3+5]),
				YsVec3(triVtxBuffer[plIdx*3+6],triVtxBuffer[plIdx*3+7],triVtxBuffer[plIdx*3+8])
			};

			double depth;
			if(YSTRUE==drawEnv.IsPolygonPicked(depth,mx,my,3,triVtPos))
			{
				if(candidateIndex<0 || depth<candidateDepth)
				{
					candidateIndex=triIdxBuffer[plIdx/3];
					candidateDepth=depth;
				}
			}
		}
	}

	z=candidateDepth;
	return candidateIndex;
}

YsArray <YsPair <int,double> >  FsGui3DMainCanvas::BoxedIndexInPointBuffer(
    int x0,int y0,int x1,int y1,
    const YsGLVertexBuffer &pointVtxBuffer,
    const YsArray <int> &pointIdxBuffer,
    const YsKeyStore &exclusion) const
{
	YsArray <YsPair <int,double> >  indexDepthPairArray;

	if(NULL!=slHd)
	{
		for(YSSIZE_T vtIdx=0; vtIdx<pointVtxBuffer.GetNumVertex(); vtIdx+=3) // 3 GL vtx per real vtx.
		{
			// Why vtIdx*3 and vtIdx/3?  See ysshellextdrawbuffer.h  It is due to unreliability of point-sprite.
			if(YSTRUE!=exclusion.IsIncluded(pointIdxBuffer[vtIdx/3]))
			{
				YsVec3 vtPos(pointVtxBuffer[vtIdx*3],pointVtxBuffer[vtIdx*3+1],pointVtxBuffer[vtIdx*3+2]);

				double z;
				if(YSTRUE==drawEnv.IsPointWithinRect(z,x0,y0,x1,y1,vtPos))
				{
					indexDepthPairArray.Increment();
					indexDepthPairArray.Last().a=pointIdxBuffer[vtIdx/3];
					indexDepthPairArray.Last().b=z;
				}
			}
		}
	}
	return indexDepthPairArray;
}

YsArray <YsPair <int,double> > FsGui3DMainCanvas::BoxedIndexInEdgeBuffer(
    int x0,int y0,int x1,int y1,const YsGLVertexBuffer &edgeVtxBuffer,const YsArray <int> &edgeIdxBuffer,const YsKeyStore *limit,const YsKeyStore *exclusion) const
{
	YsArray <YsPair <int,double> > indexDepthPairArray;

	if(NULL!=slHd)
	{
		for(YSSIZE_T edIdx=0; edIdx<=edgeVtxBuffer.GetNumVertex()-2; edIdx+=2)
		{
			if(NULL!=limit && YSTRUE!=limit->IsIncluded((YSHASHKEY)edgeIdxBuffer[edIdx/2]))
			{
				continue;
			}

			if(NULL==exclusion || YSTRUE!=exclusion->IsIncluded((YSHASHKEY)edgeIdxBuffer[edIdx/2]))
			{
				YsVec3 edVtPos[2]=
				{
					YsVec3(edgeVtxBuffer[edIdx*3  ],edgeVtxBuffer[edIdx*3+1],edgeVtxBuffer[edIdx*3+2]),
					YsVec3(edgeVtxBuffer[edIdx*3+3],edgeVtxBuffer[edIdx*3+4],edgeVtxBuffer[edIdx*3+5]),
				};

				double depth;
				if(YSTRUE==drawEnv.IsLineWithinRect(depth,x0,y0,x1,y1,edVtPos))
				{
					indexDepthPairArray.Increment();
					indexDepthPairArray.Last().a=edgeIdxBuffer[edIdx/2];
					indexDepthPairArray.Last().b=depth;
				}
			}
		}


		YsKeyStore mayBeIncluded;
		for(auto pair : indexDepthPairArray)
		{
			mayBeIncluded.AddKey(pair.a);
		}


		YsKeyStore actuallyExtendOutside;
		for(YSSIZE_T edIdx=0; edIdx<=edgeVtxBuffer.GetNumVertex()-2; edIdx+=2)
		{
			if(YSTRUE==mayBeIncluded.IsIncluded(edgeIdxBuffer[edIdx/2]))
			{
				YsVec3 edVtPos[2]=
				{
					YsVec3(edgeVtxBuffer[edIdx*3  ],edgeVtxBuffer[edIdx*3+1],edgeVtxBuffer[edIdx*3+2]),
					YsVec3(edgeVtxBuffer[edIdx*3+3],edgeVtxBuffer[edIdx*3+4],edgeVtxBuffer[edIdx*3+5]),
				};

				double depth;
				if(YSTRUE!=drawEnv.IsLineWithinRect(depth,x0,y0,x1,y1,edVtPos))
				{
					actuallyExtendOutside.AddKey(edgeIdxBuffer[edIdx/2]);
				}
			}
		}


		for(YSSIZE_T idx=indexDepthPairArray.GetN()-1; 0<=idx; --idx)
		{
			if(YSTRUE==actuallyExtendOutside.IsIncluded(indexDepthPairArray[idx].a))
			{
				indexDepthPairArray.DeleteBySwapping(idx);
			}
		}


		YsHashTable <double> idxToDepth;
		for(auto pair : indexDepthPairArray)
		{
			double depth;
			if(YSOK!=idxToDepth.FindElement(depth,pair.a))
			{
				idxToDepth.AddElement(pair.a,pair.b);
			}
			else if(depth>pair.b)
			{
				idxToDepth.UpdateElement(pair.a,pair.b);
			}
		}


		indexDepthPairArray.CleanUp();

		YsHashElementEnumHandle hd;
		if(YSOK==idxToDepth.RewindElementEnumHandle(hd))
		{
			for(;;)
			{
				YSHASHKEY key;
				double elem;
				idxToDepth.GetKey(key,hd);
				idxToDepth.GetElement(elem,hd);

				indexDepthPairArray.Increment();
				indexDepthPairArray.Last().a=(int)key;
				indexDepthPairArray.Last().b=elem;

				if(YSOK!=idxToDepth.FindNextElement(hd))
				{
					break;
				}
			}
		}
	}

	return indexDepthPairArray;
}

YsArray <YsPair <int,double> > FsGui3DMainCanvas::BoxedIndexInTriangleBuffer(
    int x0,int y0,int x1,int y1,const YsGLVertexBuffer &triVtxBuffer,const YsArray <int> &triIdxBuffer,const YsKeyStore *limit,const YsKeyStore *exclusion) const
{
	YsArray <YsPair <int,double> > indexDepthPairArray;

	if(NULL!=slHd)
	{
		for(YSSIZE_T plIdx=0; plIdx<=triVtxBuffer.GetNumVertex()-3; plIdx+=3)
		{
			if(NULL!=limit && YSTRUE!=limit->IsIncluded(triIdxBuffer[plIdx/3]))
			{
				continue;
			}

			if(NULL==exclusion || YSTRUE!=exclusion->IsIncluded(triIdxBuffer[plIdx/3]))
			{
				const YsVec3 triVtPos[3]=
				{
					YsVec3(triVtxBuffer[plIdx*3  ],triVtxBuffer[plIdx*3+1],triVtxBuffer[plIdx*3+2]),
					YsVec3(triVtxBuffer[plIdx*3+3],triVtxBuffer[plIdx*3+4],triVtxBuffer[plIdx*3+5]),
					YsVec3(triVtxBuffer[plIdx*3+6],triVtxBuffer[plIdx*3+7],triVtxBuffer[plIdx*3+8])
				};

				double depth;
				if(YSTRUE==drawEnv.IsPolygonWithinRect(depth,x0,y0,x1,y1,3,triVtPos))
				{
					indexDepthPairArray.Increment();
					indexDepthPairArray.Last().a=triIdxBuffer[plIdx/3];
					indexDepthPairArray.Last().b=depth;
				}
			}
		}


		YsKeyStore mayBeIncluded;
		for(auto pair : indexDepthPairArray)
		{
			mayBeIncluded.AddKey(pair.a);
		}


		YsKeyStore actuallyExtendOutside;
		for(YSSIZE_T plIdx=0; plIdx<=triVtxBuffer.GetNumVertex()-3; plIdx+=3)
		{
			if(YSTRUE==mayBeIncluded.IsIncluded(triIdxBuffer[plIdx/3]))
			{
				const YsVec3 triVtPos[3]=
				{
					YsVec3(triVtxBuffer[plIdx*3  ],triVtxBuffer[plIdx*3+1],triVtxBuffer[plIdx*3+2]),
					YsVec3(triVtxBuffer[plIdx*3+3],triVtxBuffer[plIdx*3+4],triVtxBuffer[plIdx*3+5]),
					YsVec3(triVtxBuffer[plIdx*3+6],triVtxBuffer[plIdx*3+7],triVtxBuffer[plIdx*3+8])
				};

				double depth;
				if(YSTRUE!=drawEnv.IsPolygonWithinRect(depth,x0,y0,x1,y1,3,triVtPos))
				{
					actuallyExtendOutside.AddKey(triIdxBuffer[plIdx/3]);
				}
			}
		}


		for(YSSIZE_T idx=indexDepthPairArray.GetN()-1; 0<=idx; --idx)
		{
			if(YSTRUE==actuallyExtendOutside.IsIncluded(indexDepthPairArray[idx].a))
			{
				indexDepthPairArray.DeleteBySwapping(idx);
			}
		}


		YsHashTable <double> idxToDepth;
		for(auto pair : indexDepthPairArray)
		{
			double depth;
			if(YSOK!=idxToDepth.FindElement(depth,pair.a))
			{
				idxToDepth.AddElement(pair.a,pair.b);
			}
			else if(depth>pair.b)
			{
				idxToDepth.UpdateElement(pair.a,pair.b);
			}
		}


		indexDepthPairArray.CleanUp();

		YsHashElementEnumHandle hd;
		if(YSOK==idxToDepth.RewindElementEnumHandle(hd))
		{
			for(;;)
			{
				YSHASHKEY key;
				double elem;
				idxToDepth.GetKey(key,hd);
				idxToDepth.GetElement(elem,hd);

				indexDepthPairArray.Increment();
				indexDepthPairArray.Last().a=(int)key;
				indexDepthPairArray.Last().b=elem;

				if(YSOK!=idxToDepth.FindNextElement(hd))
				{
					break;
				}
			}
		}
	}

	return indexDepthPairArray;
}

YsShellExtEdit *FsGui3DMainCanvas::Slhd(void)
{
	return slHd;
}

void FsGui3DMainCanvas::MakeInitialEmptyShell(void)
{
	slHd=shlGrp.CreateShell(NULL);
}

void FsGui3DMainCanvas::LoadGeneral(const wchar_t fn[],YsShellDnmContainer <YsShellExtEdit>::Node *parent)
{
	YsString fnUtf8;
	fnUtf8.EncodeUTF8 <wchar_t> (fn);

	YSFTYPE fType=YsIdentifyFileType(fnUtf8);

	switch(fType)
	{
	case YSFTYPE_SRF:
		{
			FILE *fp=YsFileIO::Fopen(fn,"r");
			if(NULL!=fp)
			{
				YsShellDnmContainer <YsShellExtEdit>::Node *newSlHd=shlGrp.CreateShell(parent);
				YsTextFileInputStream inStream(fp);

				if(YSOK==newSlHd->LoadSrf(inStream))
				{
					newSlHd->SetFileName(fn);
					File_Open_AfterOpenFile(newSlHd,fn);
				}
				else
				{
					MessageDialog(L"File read error.",L"File read error.  The file may be corrupted.");
				}
				fclose(fp);
			}
			else
			{
				MessageDialog(L"Cannot open file.",L"Cannot open file.  The file is inaccessible.");
			}
		}
		break;
	case YSFTYPE_STL:
		{
			YsShellDnmContainer <YsShellExtEdit>::Node *newSlHd=shlGrp.CreateShell(parent);
			if(YSOK==newSlHd->LoadStl(fnUtf8))
			{
				newSlHd->SetFileName(fn);
				File_Open_AfterOpenFile(newSlHd,fn);
			}
			else
			{
				MessageDialog(L"File read error.",L"File read error.  The file may be corrupted.");
			}
		}
		break;
	case YSFTYPE_DNM:
		{
			FILE *fp=YsFileIO::Fopen(fn,"r");
			if(NULL!=fp)
			{
				YsString utf8;
				utf8.EncodeUTF8 <wchar_t> (fn);

				YsTextFileInputStream inStream(fp);
				shlGrp.CleanUp();
				if(YSOK==shlGrp.LoadDnm(inStream,parent,utf8))
				{
					slHd=shlGrp.GetFirstRootNode();
					File_Open_AfterOpenFile(slHd,fn);
				}
				else
				{
					MessageDialog(FSGUI_COMMON_ERROR,FSGUI_MENU_FILE_LOADERROR);
				}
				fclose(fp);
			}
		}
		break;
	case YSFTYPE_OBJ:
		{
			FILE *fp=YsFileIO::Fopen(fn,"r");
			if(NULL!=fp)
			{
				YsShellDnmContainer <YsShellExtEdit>::Node *newSlHd=shlGrp.CreateShell(parent);
				YsTextFileInputStream inStream(fp);

				if(YSOK==newSlHd->LoadObj(inStream))
				{
					newSlHd->SetFileName(fn);
					File_Open_AfterOpenFile(newSlHd,fn);
				}
				else
				{
					MessageDialog(L"File read error.",L"File read error.  The file may be corrupted.");
				}
				fclose(fp);
			}
			else
			{
				MessageDialog(L"File read error.",L"File read error.  The file may be corrupted.");
			}
		}
		break;
	default:
		MessageDialog(L"Error.",L"Unsupported file type.");
		break;
	}
}

void FsGui3DMainCanvas::SaveGeneral(const YsShellExtEdit &shl,const wchar_t fn[])
{
	YsString fnUtf8;
	fnUtf8.EncodeUTF8 <wchar_t> (fn);

	YSFTYPE fType=YsIdentifyFileType(fnUtf8);

	YsWString ful,pth,fil;
	ful.Set(fn);
	ful.SeparatePathFile(pth,fil);

	YsWString savedMsg;
	savedMsg.Append(L"Saved:");
	savedMsg.Append(fil);
	savedMsg.Append(L"\n");
	savedMsg.Append(L"Path: ");
	savedMsg.Append(pth);

	switch(fType)
	{
	case YSFTYPE_SRF:
		{
			FILE *fp=YsFileIO::Fopen(fn,"w");
			if(NULL!=fp)
			{
				YsTextFileOutputStream outStream(fp);

				YsShellExtWriter writer;
				writer.SaveSrf(outStream,(const YsShellExt &)shl);
				shl.SetFileName(fn);
				shl.Saved();
				lastAccessedFileName.Set(fn);

				fclose(fp);

				MessageDialog(FSGUI_DLG_SAVED_TITLE,savedMsg);
			}
			else
			{
				MessageDialog(FSGUI_COMMON_ERROR,L"Cannot open the file in write-mode.");
			}
		}
		break;
	case YSFTYPE_OBJ:
		{
			FILE *fp=YsFileIO::Fopen(fn,"w");
			if(NULL!=fp)
			{
				YsTextFileOutputStream outStream(fp);

				YsShellExtObjWriter writer;
				YsShellExtObjWriter::WriteOption option;
				writer.WriteObj(outStream,(const YsShellExt &)shl,option);
				shl.SetFileName(fn);
				shl.Saved();
				lastAccessedFileName.Set(fn);

				fclose(fp);

				MessageDialog(FSGUI_DLG_SAVED_TITLE,savedMsg);
			}
			else
			{
				MessageDialog(FSGUI_COMMON_ERROR,L"Cannot open the file in write-mode.");
			}
		}
		break;
	case YSFTYPE_STL:
		{
			for(auto plHd : shl.AllPolygon())
			{
				if(3!=shl.GetNumVertexOfPolygon(plHd))
				{
					MessageDialog(FSGUI_COMMON_ERROR,
					              L"The model includes at least one non-triangular polygon.\n"
					              L"STL can only save triangles.\n"
					              L"Please use Global->Triangulate All to make all triangles.");
					return;
				}
			}

			FILE *fp=YsFileIO::Fopen(fn,"w");
			if(NULL!=fp)
			{
				int nIgnored;
				shl.SaveBinStl(nIgnored,fp,"STL generated by PolygonCrest editor.");
				fclose(fp);

				shl.SetFileName(fn);
				MessageDialog(FSGUI_DLG_SAVED_TITLE,savedMsg);
			}
			else
			{
				MessageDialog(FSGUI_COMMON_ERROR,L"Cannot open the file in write-mode.");
			}
		}
		break;
	default:
		MessageDialog(FSGUI_COMMON_ERROR,
		              L"The model cannot be saved in this file type.\n"
		              L"Please use SaveAs and specify a different file extension.");
		break;
	}
}

YsWString FsGui3DMainCanvas::GetDataDir(void) const
{
	if(0<lastAccessedFileName.Strlen())
	{
		YsWString pth,fil;
		lastAccessedFileName.SeparatePathFile(pth,fil);
		return pth;
	}
	else if(NULL!=config && 0<config->defDataDir.Strlen())
	{
		return config->defDataDir;
	}
	else
	{
		return YsWString(L".");
	}
}

void FsGui3DMainCanvas::AddShortCut(int fskey,YSBOOL shift,YSBOOL ctrl,void (*callBackFunc)(void *,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *))
{
	for(YSSIZE_T idx=0; idx<shortCut.GetN(); ++idx)
	{
		if(shortCut[idx].fskey==fskey && shortCut[idx].shift==shift && shortCut[idx].ctrl==ctrl)
		{
			YsPrintf("Internal Error!\n");
			YsPrintf("  Shortcut already defined!\n");
			exit(1);
		}
	}

	shortCut.Increment();
	shortCut.GetEnd().callBackFunc=callBackFunc;
	shortCut.GetEnd().fskey=fskey;
	shortCut.GetEnd().shift=shift;
	shortCut.GetEnd().ctrl=ctrl;
}

void FsGui3DMainCanvas::RefreshRecentlyUsedFileList(void)
{
	YsWString recentFn=fileAndDirName->GetRecentFileListFileName();
	FILE *fp=YsFileIO::Fopen(recentFn,"r");
	if(NULL!=fp)
	{
		YsTextFileInputStream inStream(fp);
		recent.Open(inStream);
		fclose(fp);

		recent.PopulateMenu(*fileRecent,16,File_Recent,this);
	}
}

void FsGui3DMainCanvas::AddRecentlyUsedFile(const wchar_t wfn[])
{
	recent.AddFile(wfn);
	recent.PopulateMenu(*fileRecent,16,File_Recent,this);

	YsWString recentFn=fileAndDirName->GetRecentFileListFileName();
	FILE *fp=YsFileIO::Fopen(recentFn,"w");
	if(NULL!=fp)
	{
		YsTextFileOutputStream outStream(fp);
		recent.Save(outStream,16);
		fclose(fp);
	}
}


////////////////////////////////////////////////////////////



YSBOOL FsGui3dAppMustTerminate(void)
{
	return appMustTerminate;
}

void FsGui3DSetAppMustTerminate(YSBOOL sw)
{
	appMustTerminate=sw;
}

int FsGui3dAppSleepIntervalInMS(void)
{
	return appSleepInterval;
}

static void FsGui3DAppLocalIntervalHook(void *)
{
	FsGui3DAppIntervalCallBack();
}

class FsGui3DAppCommandParameterInfo
{
public:
	YSBOOL forceEnglishMode;
	YSBOOL peacefullyExit;

	FsGui3DAppCommandParameterInfo();
	YSRESULT RecognizeCommandParameter(YsArray <char *> &args);
};

FsGui3DAppCommandParameterInfo::FsGui3DAppCommandParameterInfo()
{
	forceEnglishMode=YSFALSE;
	peacefullyExit=YSFALSE;
}

YSRESULT FsGui3DAppCommandParameterInfo::RecognizeCommandParameter(YsArray <char *> &args)
{
	YSSIZE_T idx=0;
	while(idx<args.GetN())
	{
		if(0==strcmp(args[idx],"-en"))
		{
			forceEnglishMode=YSTRUE;
			args.Delete(idx);
		}
		else if(0==strcmp(args[idx],"-h"))
		{
			YsPrintf("Command Parameters:\n");
			YsPrintf("  -h Print help.\n");
			YsPrintf("  -en Force English mode.\n");
			peacefullyExit=YSTRUE;
			args.Delete(idx);
		}
		else
		{
			++idx;
		}
	}
	return YSOK;
}

void FsGui3DAppInitialize(int ac,char *av[])
{
	YsArray <char *> args;
	for(int i=0; i<ac; ++i)
	{
		args.Append(av[i]);
	}

	FsGui3DAppCommandParameterInfo cpmi;
	if(YSOK!=cpmi.RecognizeCommandParameter(args))
	{
		printf("Parameter Error!\n");
		exit(1);
	}
	if(YSTRUE==cpmi.peacefullyExit)
	{
		exit(1);
	}

	YsWString progBaseDir;
	if(YSOK==YsSpecialPath::GetProgramBaseDir(progBaseDir))
	{
		YsLocale locale;
		const char *langStr=locale.GetLanguagePart();

		if(YSTRUE==cpmi.forceEnglishMode)
		{
			langStr="en";
		}

		if(NULL!=langStr)
		{
			YsWString langWStr;
			langWStr.SetUTF8String(langStr);

			YsWString langPath;
			langPath.MakeFullPathName(progBaseDir,L"language");

			YsWString langFileName;
			langFileName.MakeFullPathName(langPath,langWStr);

			langFileName.ReplaceExtension(L".uitxt");

			printf("%ls\n",langFileName.Txt());

			FILE *fp=YsFileIO::Fopen(langFileName,"r");
			if(NULL!=fp)
			{
				fsGuiTextResource.LoadFile(fp);
				fclose(fp);
			}
		}
	}



	appMainCanvas=new FsGui3DMainCanvas;
	YsGLSLCreateSharedRenderer();

	appMainCanvas->drawEnv.SetTargetBoundingBox(-10.0*YsXYZ(),10.0*YsXYZ());
	appMainCanvas->drawEnv.SetViewTarget(YsOrigin());
	appMainCanvas->drawEnv.SetViewDistance(15.0);

	appMainCanvas->MakeInitialEmptyShell();

	appMainCanvas->colorPaletteDlg->MakeDialog(appMainCanvas);
	appMainCanvas->AddDialog(appMainCanvas->colorPaletteDlg);

	FsRegisterIntervalCallBack(FsGui3DAppLocalIntervalHook,NULL);

	appMainCanvas->fileAndDirName->GetOptionFileName();

	appMainCanvas->Option_ReapplyConfig();

	if(YSOPENGL==YsCoordSysModel)
	{
		appMainCanvas->drawEnv.SetViewDirection(-YsXYZ());
	}
	else
	{
		appMainCanvas->drawEnv.SetViewDirection(YsVec3(1.0,-1.0,-1.0));
	}



#ifndef __APPLE__
	// There's no such thing called command parameter in Mac OSX
	if(2<=args.GetN())
	{
		YsWString wfn;
		wfn.SetUTF8String(args[1]);

		YsWString pth,fil;
		wfn.SeparatePathFile(pth,fil);

		if(0==pth.Strlen())
		{
			wfn.MakeFullPathName(L".",fil);
		}

		appMainCanvas->LoadGeneral(wfn,NULL);
	}
#endif
}

void FsGui3DAppIntervalCallBack(void)
{
	appMainCanvas->OnInterval();
}

void FsGui3dAppCleanUp(void)
{
	YsGLSLDeleteSharedRenderer();
	delete appMainCanvas;
}

