/* ////////////////////////////////////////////////////////////

  File Name: ysshellext_boundaryutil.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellext_boundaryutil.h"
#include "ysshellext_trackingutil.h"

YsShellExt_BoundaryInfo::YsShellExt_BoundaryInfo()
{
	CleanUp();
}

YsShellExt_BoundaryInfo::~YsShellExt_BoundaryInfo()
{
	CleanUp();
}

void YsShellExt_BoundaryInfo::CleanUp(void)
{
	srcEdVtHd.CleanUp();
	edgeToPlgHash.CleanUp();
	visited.CleanUp();
	nonDuplicateVtHdArray.CleanUp();
	contourCache.CleanUp();
}


void YsShellExt_BoundaryInfo::MakeInfo(
    const YsShell &shl,
    YSSIZE_T nPl,const YsShellPolygonHandle plHdPtr[])
{
	visited.SetShell((const YsShell &)shl);

	auto plHdArray=YsMakeArrayMask(nPl,plHdPtr);

	for(auto plHd : plHdArray)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetVertexListOfPolygon(plVtHd,plHd);

		for(int edIdx=0; edIdx<plVtHd.GetN(); ++edIdx)
		{
			const unsigned int edVtKey[2]=
			{
				shl.GetSearchKey(plVtHd[edIdx]),
				shl.GetSearchKey(plVtHd.GetCyclic(edIdx+1))
			};
			edgeToPlgHash.AddElement(2,edVtKey,plHd);
		}
	}

	for(auto plHd : plHdArray)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetVertexListOfPolygon(plVtHd,plHd);

		for(int edIdx=0; edIdx<plVtHd.GetN(); ++edIdx)
		{
			const unsigned int edVtKey[2]=
			{
				shl.GetSearchKey(plVtHd[edIdx]),
				shl.GetSearchKey(plVtHd.GetCyclic(edIdx+1))
			};

			int nEdPl;
			const YsShellPolygonHandle *edPlHd;
			if(YSOK==edgeToPlgHash.FindElement(nEdPl,edPlHd,2,edVtKey) && 1==nEdPl)
			{
				const YsShellVertexHandle edVtHd[2]=
				{
					plVtHd[edIdx],
                    plVtHd.GetCyclic(edIdx+1)
				};

				srcEdVtHd.Append(edVtHd[0]);
				srcEdVtHd.Append(edVtHd[1]);
				visited.AddEdge(edVtHd[0],edVtHd[1]);
			}
		}
	}
}

void YsShellExt_BoundaryInfo::MakeVertexConnectionTable(
		YsFixedLengthToMultiHashTable <YsShellVertexHandle,1,2> &vtKeyToConnVtHd,
		const YsShell &shl) const
{
	vtKeyToConnVtHd.PrepareTable();
	vtKeyToConnVtHd.EnableAutoResizing();

	for(YSSIZE_T idx=0; idx<=srcEdVtHd.GetN()-2; idx+=2)
	{
		const unsigned vtKey[2]=
		{
			shl.GetSearchKey(srcEdVtHd[idx]),
			shl.GetSearchKey(srcEdVtHd[idx+1])
		};

		vtKeyToConnVtHd.AddElement(1,vtKey+0,srcEdVtHd[idx+1]);
		vtKeyToConnVtHd.AddElement(1,vtKey+1,srcEdVtHd[idx+0]);
	}
}

void YsShellExt_BoundaryInfo::CacheNonDuplicateVtHdArray(void)
{
	nonDuplicateVtHdArray=srcEdVtHd;
	YsRemoveDuplicateInUnorderedArray(nonDuplicateVtHdArray);
}

YsVec3 YsShellExt_BoundaryInfo::GetCenter(const YsShellExt &shl) const
{
	YsVec3 cen=YsOrigin();
	if(0<nonDuplicateVtHdArray.GetN())
	{
		for(YSSIZE_T vtIdx=0; vtIdx<nonDuplicateVtHdArray.GetN(); ++vtIdx)
		{
			YsVec3 pos;
			shl.GetVertexPosition(pos,nonDuplicateVtHdArray[vtIdx]);
			cen+=pos;
		}
		cen/=(double)nonDuplicateVtHdArray.GetN();
	}
	return cen;
}

const YsArray <YsShellVertexHandle> &YsShellExt_BoundaryInfo::GetVertexAll(void) const
{
	return nonDuplicateVtHdArray;
}

YSRESULT YsShellExt_BoundaryInfo::CacheContour(const YsShell &shl)
{
	YsFixedLengthToMultiHashTable <YsShellVertexHandle,1,2> connection;

	MakeVertexConnectionTable(connection,shl);
	CacheNonDuplicateVtHdArray();

	contourCache.CleanUp();

	// First, start from the vertices with 3 or more connections.
	YsArray <YsShellVertexHandle> moreThanTwo,twoConnection;
	for(auto vtHd : nonDuplicateVtHdArray)
	{
		YSSIZE_T nConnVtHd;
		const YsShellVertexHandle *connVtHd;
		const unsigned int vtKey[1]={shl.GetSearchKey(vtHd)};
		if(YSOK==connection.FindElement(nConnVtHd,connVtHd,1,vtKey) && 3<=nConnVtHd)
		{
			moreThanTwo.Append(vtHd);
		}
		else if(2==nConnVtHd)
		{
			twoConnection.Append(vtHd);
		}
		else if(1==nConnVtHd)
		{
			contourCache.CleanUp();
			return YSERR;
		}
	}

	YsTwoKeyStore usedEdgePiece;
	for(auto vtHd : moreThanTwo)
	{
		YSSIZE_T nConnVtHd;
		const YsShellVertexHandle *connVtHd;
		const unsigned int vtKey[1]={shl.GetSearchKey(vtHd)};
		if(YSOK==connection.FindElement(nConnVtHd,connVtHd,1,vtKey))
		{
			for(int i=0; i<nConnVtHd; ++i)
			{
				if(YSOK!=TrackAndAddLoop(usedEdgePiece,shl,vtHd,connVtHd[i],connection))
				{
					contourCache.CleanUp();
					return YSERR;
				}
			}
		}
	}

	for(auto vtHd : twoConnection)
	{
		YSSIZE_T nConnVtHd;
		const YsShellVertexHandle *connVtHd;
		const unsigned int vtKey[1]={shl.GetSearchKey(vtHd)};
		if(YSOK==connection.FindElement(nConnVtHd,connVtHd,1,vtKey))
		{
			for(int i=0; i<nConnVtHd; ++i)
			{
				if(YSOK!=TrackAndAddLoop(usedEdgePiece,shl,vtHd,connVtHd[i],connection))
				{
					contourCache.CleanUp();
					return YSERR;
				}
			}
		}
	}

	for(YSSIZE_T idx=0; idx<=srcEdVtHd.GetN()-2; idx+=2)
	{
		const unsigned vtKey[2]=
		{
			shl.GetSearchKey(srcEdVtHd[idx]),
			shl.GetSearchKey(srcEdVtHd[idx+1])
		};
		if(YSTRUE!=usedEdgePiece.IsIncluded(vtKey[0],vtKey[1]))
		{
			contourCache.CleanUp();
			return YSERR;
		}
	}

	return YSOK;
}

YSRESULT YsShellExt_BoundaryInfo::TrackAndAddLoop(
    YsTwoKeyStore &usedEdgePiece,
    const YsShell &shl,YsShellVertexHandle fromVtHd,YsShellVertexHandle toVtHd,
    const YsFixedLengthToMultiHashTable <YsShellVertexHandle,1,2> &connection)
{
	if(YSTRUE==usedEdgePiece.IsIncluded(shl.GetSearchKey(fromVtHd),shl.GetSearchKey(toVtHd)))
	{
		return YSOK;
	}

	YsArray <YsShellVertexHandle> path=YsShellExt_TrackingUtil::TrackVertexConnection(shl,fromVtHd,toVtHd,connection);
	if(2<path.GetN() && path[0]==path.Last())
	{
		contourCache.Increment();
		contourCache.Last().Initialize();
		contourCache.Last().vtHdArray=path;

		for(YSSIZE_T vtIdx=0; vtIdx<path.GetN()-1; ++vtIdx)
		{
			usedEdgePiece.AddKey(shl.GetSearchKey(path[vtIdx]),shl.GetSearchKey(path[vtIdx+1]));
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSSIZE_T YsShellExt_BoundaryInfo::GetNumContour(void) const
{
	return contourCache.GetN();
}

YSRESULT YsShellExt_BoundaryInfo::GetContour(YSSIZE_T &nVt,const YsShellVertexHandle *&vtHdArray,YSSIZE_T idx) const
{
	if(YSTRUE==contourCache.IsInRange(idx))
	{
		nVt=contourCache[idx].vtHdArray.GetN();
		vtHdArray=contourCache[idx].vtHdArray;
		return YSOK;
	}
	nVt=0;
	vtHdArray=NULL;
	return YSERR;
}

YsArray <YsArray <YsShellVertexHandle> > YsShellExt_BoundaryInfo::GetContourAll(void) const
{
	YsArray <YsArray <YsShellVertexHandle> > allContour(GetNumContour(),NULL);
	for(YSSIZE_T idx=0; idx<GetNumContour(); ++idx)
	{
		GetContour(allContour[idx],idx);
	}
	return allContour;
}

void YsShellExt_BoundaryInfo::GetEdgeUsageMinMax(int &min,int &max,const YsShellExt &shl) const
{
	if(2<=srcEdVtHd.GetN())
	{
		min=shl.GetNumPolygonUsingEdge(srcEdVtHd[0],srcEdVtHd[1]);
		max=min;
		for(YSSIZE_T edIdx=2; edIdx<=srcEdVtHd.GetN()-2; edIdx+=2)
		{
			const int n=shl.GetNumPolygonUsingEdge(srcEdVtHd[edIdx],srcEdVtHd[edIdx+1]);
			min=YsSmaller(min,n);
			max=YsGreater(max,n);
		}
	}
	else
	{
		min=0;
		max=0;
	}
}

YSSIZE_T YsShellExt_BoundaryInfo::GetNumEdgePiece(void) const
{
	return srcEdVtHd.GetN()/2;
}

YSRESULT YsShellExt_BoundaryInfo::GetEdge(YsShellVertexHandle edVtHd[2],YSSIZE_T edIdx) const
{
	if(0<=edIdx && edIdx*2+1<srcEdVtHd.GetN())
	{
		edVtHd[0]=srcEdVtHd[edIdx*2];
		edVtHd[1]=srcEdVtHd[edIdx*2+1];
		return YSOK;
	}
	return YSERR;
}

