#ifndef YSSHELLEXT_GEOMCALC_IS_INCLUDED
#define YSSHELLEXT_GEOMCALC_IS_INCLUDED
/* { */

#include <ysclass.h>

/*! Calculate dihedral angle between two polygons. 
    These functions calculate normals based on vertex locations, not cached normal vectors. */
const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellPolygonHandle plHd[2]);
const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,YsShellPolygonHandle plHd0,YsShellPolygonHandle plHd1);

/*! Calculate dihedral angle across an edge.
    If the edge is not used by two polygons, this function returns 0.0 */
const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellVertexHandle edVtHd[2]);
const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1);
const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellPolygonHandle plHd,int nThEdgeZeroBased);

/*! Calculate total dihedral angle of edges:
    This function internally uses YsShellExt_CalculateDihedralAngle function.
           edVtHd[0]
    +--(1)--+
    |      /|
    |     / |
   (2)   /  |
    |  (5)  |
    |  /   (4)
    | /     |
    |/      |
    +--(3)--+
  edVtHd[1] */
const double YsShellExt_CalculateTotalDihedralAngleAroundEdge(const class YsShellExt &shl,const YsShellVertexHandle edVtHd[2]);
const double YsShellExt_CalculateTotalDihedralAngleAroundEdge(const class YsShellExt &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1);

/*! This function calculates average normal of the given polygons.  It does not use cached normal.
    It only rely on vertex positions. 
    It is not orientation sensitive.  The returned normal may be reverse of the normal that can be calculated based on the polygon orientation. */
const YsVec3 YsShell_CalculatePolygonAverageNormal(const YsShell &shl,YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[]);
template <const int N>
const YsVec3 YsShell_CalculatePolygonAverageNormal(const YsShell &shl,const YsArray <YsShellPolygonHandle,N> &plHdArray)
{
	return YsShell_CalculatePolygonAverageNormal(shl,plHdArray.GetN(),plHdArray);
}


/* } */
#endif
