#ifndef YSSYSTEMFONT_IS_INCLUDED
#define YSSYSTEMFONT_IS_INCLUDED
/* { */

#ifndef YSRESULT_IS_DEFINED
#define YSRESULT_IS_DEFINED
typedef enum
{
	YSERR,
	YSOK
} YSRESULT;
#endif

#ifndef YSBOOL_IS_DEFINED
#define YSBOOL_IS_DEFINED
typedef enum
{
	YSFALSE,
	YSTRUE,
	YSTFUNKNOWN
} YSBOOL;
#endif

class YsSystemFontTextBitmap
{
friend class YsSystemFontCache;

private:
	unsigned int wid,hei;
	unsigned int bytePerLine;
	unsigned int bitPerPixel;
	unsigned char *dat;
public:
	inline YsSystemFontTextBitmap();
	inline ~YsSystemFontTextBitmap();
	inline void CleanUp(void);
	inline unsigned int Width(void) const;
	inline unsigned int Height(void) const;
	inline unsigned int BytePerLine(void) const;
	inline unsigned int BitPerPixel(void) const;
	inline const unsigned char *Bitmap(void) const;
	inline unsigned char *TransferBitmap(void);

private:
	inline void SetBitmap(unsigned int wid,unsigned int hei,unsigned int bytePerLine,unsigned int bitPerPixel,unsigned char *dat);
};

inline YsSystemFontTextBitmap::YsSystemFontTextBitmap()
{
	dat=NULL;
	CleanUp();
}

inline YsSystemFontTextBitmap::~YsSystemFontTextBitmap()
{
	CleanUp();
}

inline void YsSystemFontTextBitmap::CleanUp(void)
{
	if(NULL!=dat)
	{
		delete [] dat;
		dat=NULL;
	}
	wid=0;
	hei=0;
	bytePerLine=0;
	bitPerPixel=0;
}

inline unsigned int YsSystemFontTextBitmap::Width(void) const
{
	return wid;
}

inline unsigned int YsSystemFontTextBitmap::Height(void) const
{
	return hei;
}

inline unsigned int YsSystemFontTextBitmap::BytePerLine(void) const
{
	return bytePerLine;
}

inline unsigned int YsSystemFontTextBitmap::BitPerPixel(void) const
{
	return bitPerPixel;
}

inline const unsigned char *YsSystemFontTextBitmap::Bitmap(void) const
{
	return dat;
}

inline unsigned char *YsSystemFontTextBitmap::TransferBitmap(void)
{
	unsigned char *ret=dat;
	dat=NULL;
	CleanUp();
	return ret;
}

inline void YsSystemFontTextBitmap::SetBitmap(unsigned int wid,unsigned int hei,unsigned int bytePerLine,unsigned int bitPerPixel,unsigned char *dat)
{
	CleanUp();
	this->wid=wid;
	this->hei=hei;
	this->bytePerLine=bytePerLine;
	this->bitPerPixel=bitPerPixel;
	this->dat=dat;
}



class YsSystemFontCache
{
private:
	class YsSystemFontInternal *internal;

public:
	YsSystemFontCache();
	~YsSystemFontCache();

	YSRESULT RequestDefaultFont(void);
	YSRESULT RequestDefaultFontWithHeight(int height);

	inline YSRESULT MakeRGBABitmap(YsSystemFontTextBitmap &bmp,const wchar_t wStr[],unsigned int fgCol,const unsigned int bgCol,YSBOOL reverse) const;

	YSRESULT MakeSingleBitBitmap(YsSystemFontTextBitmap &bmp,const wchar_t wStr[],YSBOOL reverse) const;
	YSRESULT MakeRGBABitmap(YsSystemFontTextBitmap &bmp,const wchar_t wStr[],const unsigned char fgCol[3],const unsigned char bgCol[3],YSBOOL reverse) const;
	YSRESULT MakeGrayScaleAndAlphaBitmap(YsSystemFontTextBitmap &bmp,const wchar_t wStr[],unsigned char fgCol,unsigned char bgCol,YSBOOL reverse) const;
	YSRESULT GetTightBitmapSize(int &wid,int &hei,const wchar_t wStr[]) const;
};

inline YSRESULT YsSystemFontCache::MakeRGBABitmap(YsSystemFontTextBitmap &bmp,const wchar_t wStr[],unsigned int fgColI,const unsigned int bgColI,YSBOOL reverse) const
{
	const unsigned char fgCol[3]={(unsigned char)((fgColI>>16)&255),(unsigned char)((fgColI>>8)&255),(unsigned char)(fgColI&255)};
	const unsigned char bgCol[3]={(unsigned char)((bgColI>>16)&255),(unsigned char)((bgColI>>8)&255),(unsigned char)(bgColI&255)};
	return MakeRGBABitmap(bmp,wStr,fgCol,bgCol,reverse);
}

/* } */
#endif
